import os

import yaml
from sqlalchemy.engine.url import URL

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
home_directory = os.path.join(os.path.expanduser('~'), '.edam')
test_resources = os.path.join(ROOT_DIR, '../', 'tests', 'resources')
settings = os.path.join(home_directory, 'settings.yaml')

with open(settings, 'r') as stream:
    try:
        settings_content = yaml.load(stream, Loader=yaml.FullLoader)
    except yaml.YAMLError as exc:
        raise exc


def safe_return(section, fields_as_dict):
    temp_dict = dict()
    for key, default_value in fields_as_dict.items():
        try:
            value = settings_content[section][key]
        except:
            value = default_value
        temp_dict[key] = value
    
    if section == 'DATABASE':
        if temp_dict['drivername'] in ['sqlite']:
            temp_dict['database'] = os.path.join(home_directory, temp_dict['database'])
    
    return temp_dict


database_fields = {'drivername': 'sqlite', 'database': 'edam.db'}
DATABASE = safe_return('DATABASE', database_fields)

server_fields = {'host': '127.0.0.1', 'port': '5000', 'debug': False}
SERVER = safe_return('SERVER', server_fields)
ogc_sos_fields = {'title': 'EDAM',
                  'description': 'EDAM OGC Sensor Observation Service',
                  'name': 'First name Last name',
                  'provider_name': 'Wageningen University',
                  'voice': '00302541-079-516',
                  'fax': '00302541-079-516',
                  'delivery_point': 'WUR',
                  'city': 'Wageningen',
                  'administration_area': 'Netherlands',
                  'postal_code': '',
                  'country': 'Netherlands'}

OGC_SOS_CONFIGURATION = safe_return('OGC SOS', ogc_sos_fields)
database_url = URL(**DATABASE)
database_type = DATABASE['drivername']

if __name__ == "__main__":
    print(test_resources)



import io
import logging
import re

from edam.reader.utilities import remove_template_placeholders_from_string


class Preprocess(object):
    """
    This class will produce 4 files of 2 types (data and preamble) if applicable.
    It identifies the case an input document has preamble, and created two separate
    files (_data, _preamble). Given that input and .tmpl documents have the same
    structure, this "split" process is applied to both input and .tmpl documents
    (thus, 4 documents in total).
    """
    
    def __init__(self, input_file, template_file):
        """
        
        :param input_file: Initial input file path (eg ~/inputs/Agmip.csv)
        :param template_file: Initial template file path (eg ~/templates/Agmip.tmpl)
        """
        self.preprocess_logger = logging.getLogger('edam.reader.Preprocess.Preprocess')
        self.input_file = input_file
        self.input_file.seek(0)
        self.template_file = template_file
        self.template_file.seek(0)
        
        self.header = ""
        self.extract_header_from_template()
        
        self.new_input_file_data = None
        self.new_input_file_preamble = None
        self.new_template_file_data = None
        self.new_template_file_preamble = None
        
        if self.header != "":
            self.new_template_file_data, self.new_template_file_preamble, \
            self.new_input_file_data, self.new_input_file_preamble = \
                self.split_preamble_from_data(self.template_file, self.input_file)
    
    def extract_header_from_template(self):
        var_header_from_template_file = r"((.*\n){1}){%\s?for.*%}"
        regex_header_from_template_file = re.compile(var_header_from_template_file)
        template_text = self.template_file.read()
        self.template_file.seek(0)
        matches = re.findall(regex_header_from_template_file, template_text)
        if matches:
            self.header = matches[0][0].strip("\r\n")
            self.preprocess_logger.debug('Header is: %s' % self.header)
        else:
            self.preprocess_logger.error('Template does not have header line')
            self.preprocess_logger.error('Maybe it is empty')
            self.preprocess_logger.error('Please check: %s' % self.template_file)
            raise SystemExit("Empty header line")
        
        self.template_file.seek(0)
    
    def split_preamble_from_data(self, template_file, input_file):
        # eg. py, pdf
        
        # Consider a header which contains placeholders
        # e.g "Speed of maximum wind gust (km/h)","Time of maximum wind gust",{{temp.timestamp.hour=9}}
        
        # Above header differs from template file to input file.
        # So we have two types of headers.
        # 1. template_header which is the above and
        # 2. input_header which is similar to template_header (in place of the placeholder we have something different
        self.template_header = self.header.strip('\n')
        header_without_placeholders = remove_template_placeholders_from_string(self.header).strip('\n')
        input_text = input_file
        for line in input_text.readlines():
            cleaned_line = line.strip('\n')
            if header_without_placeholders in cleaned_line or header_without_placeholders in cleaned_line.replace(' ',
                                                                                                                  ''):
                self.input_header = cleaned_line
                break
        
        # template_file_to_be_parsed_as_list_by_period = os.path.splitext(template_file)
        # input_file_to_be_parsed_as_list_by_period = os.path.splitext(input_file)
        
        template_data_file_text = template_file
        input_data_file_text = input_file
        
        template_file_text = template_file.read()
        template_file.seek(0)
        input_file.seek(0)
        input_file_text = input_file.read()
        input_file.seek(0)
        
        template_lines_before, _, template_lines_after = template_file_text.partition(self.template_header)
        input_lines_text_before, _, input_lines_text_after = input_file_text.partition(self.input_header)
        # TODO: I should probably insert an else statement here.
        # TODO: If there is no match, I could just rename the file to sth_data.sth
        
        # We can safely presume that the same condition verifies input text
        
        template_preamble_file_text = io.StringIO()
        input_preamble_file_text = io.StringIO()
        if template_lines_before != "":
            # Find preamble text and write file
            template_preamble_text = template_lines_before
            input_preamble_text = input_lines_text_before
            template_preamble_file_text.write(template_preamble_text)
            input_preamble_file_text.write(input_preamble_text)
            
            # Remove preamble text and save only data
            # We will use re.sub function
            # Preamble will be substituted by the header
            
            template_data_text = self.header + template_lines_after.replace(template_preamble_text, '')
            input_data_text = self.header + input_lines_text_after.replace(input_preamble_text, '')
            
            template_data_file_text = io.StringIO()
            input_data_file_text = io.StringIO()
            
            template_data_file_text.write(template_data_text)
            input_data_file_text.write(input_data_text)
        # That means document does not have a preamble
        template_data_file_text.seek(0)
        template_preamble_file_text.seek(0)
        input_data_file_text.seek(0)
        input_preamble_file_text.seek(0)
        
        return template_data_file_text, template_preamble_file_text, input_data_file_text, input_preamble_file_text


if __name__ == "__main__":
    test = Preprocess("inputs/Agmip.csv", "templates/Agmip.tmpl")



from edam.viewer.app.views import app
from edam.settings import SERVER

app.jinja_env.auto_reload = True
app.config['TEMPLATES_AUTO_RELOAD'] = True
app.run(host=SERVER['host'], port=SERVER['port'], debug=SERVER['debug'])



from edam.viewer.app.views import app
from edam.settings import SERVER


def run():
    app.jinja_env.auto_reload = True
    app.config['TEMPLATES_AUTO_RELOAD'] = True
    app.run(host=SERVER['host'], port=SERVER['port'], debug=SERVER['debug'])


if __name__ == "__main__":
    run()



from edam.reader.manage import DatabaseHandler
from edam.settings import OGC_SOS_CONFIGURATION
from edam.viewer.app.manage import Measurement


class OGC_SOS:
    def __init__(self, request, procedure=None, offering=None, eventTime=None, observedProperty=None,
                 page=None):
        self.available_requests = ['GetCapabilities', 'DescribeSensor', 'GetObservation']
        self.offering = offering
        self.eventTime = eventTime
        self.observedProperty = observedProperty
        self.info = OGC_SOS_CONFIGURATION
        self.keywords = list()
        self.stations = list()
        self.results = list()
        self.metadata = list()
        self.procedure = procedure
        self.sensor = None
        self.page = page
        self.template = None
        
        self.exception = False
        self.helper_object = None
        
        self.exceptionDetails = {}
        
        self.data = DatabaseHandler()
        
        if request in self.available_requests:
            self.request = request
            self.determine_request()
    
    def determine_request(self):
        if self.request == "GetCapabilities":
            self.find_keywords()
            self.find_stations()
            self.find_metadata()
            self.template = "sos/GetCapabilities.xml"
        elif self.request == "DescribeSensor":
            # self.procedure is like: station_name:sensor_name:template_id
            
            try:
                station_id, sensor_id, template_id = self.procedure.split(':')
                exists = self.data.get_helper_for_describe_sensor(station_id=station_id, sensor_id=sensor_id,
                                                                  template_id=template_id)
                if exists:
                    
                    self.sensor = exists
                    self.template = "sos/DescribeSensor.xml"
                else:
                    self.template = "sos/DescribeSensorException.xml"
            except:
                self.template = "sos/DescribeSensorException.xml"
        elif self.request == "GetObservation":
            try:
                station_id, sensor_id, template_id = self.procedure.split(':')
                exists = self.data.get_helper_for_describe_sensor(station_id=station_id, sensor_id=sensor_id,
                                                                  template_id=template_id)
                if exists:
                    # from_time, to_time = self.eventTime.split('/')
                    # from_time = pd.to_datetime(from_time)
                    # to_time = pd.to_datetime(to_time)
                    
                    self.helper_object = exists
                    results = self.data.get_observations_by_helper_id(self.helper_object.id)
                    for row in results:
                        self.results.append(Measurement(value=row.value, timestamp=row.timestamp,
                                                        observable=self.helper_object.observable,
                                                        uom=self.helper_object.uom,
                                                        station=self.helper_object.station,
                                                        helper=self.helper_object))
                    # self.results = [Measurement(value=row.value, timestamp=row.timestamp) for row in results]
                    
                    self.template = "sos/GetObservation.xml"
                else:
                    self.template = "sos/GetObservationException.xml"
            except Exception as inst:
                print(inst)
                self.template = "sos/GetObservationException.xml"
    
    def find_keywords(self):
        [self.keywords.append(quantity.name) for quantity in self.data.get_all_observables()]
    
    def find_stations(self):
        [self.stations.append(station) for station in self.data.get_all_stations()]
    
    def find_metadata(self):
        [self.metadata.append(helper) for helper in self.data.get_all_helper_observable_ids()]



import json
import logging

from sqlalchemy import Column, Integer, String, Boolean, DateTime, ForeignKey, Float
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

Base = declarative_base()

module_logger = logging.getLogger('edam.reader.models')


class Station(Base):
    __tablename__ = "Station"
    """
    :var id: A unique id among other Stations
    :var name: The name of the Station
    :var mobile: Boolean, i.e. True of False
    :var polygon: :If mobile=True, then a polygon is given
    :var latitude:
    :var longitude:
    """
    
    id = Column(Integer, primary_key=True)
    name = Column(String(80))
    mobile = Column(Boolean)
    location = Column(String(200))
    latitude = Column(Float)
    longitude = Column(Float)
    region = Column(String(200))
    license = Column(String(100))
    url = Column(String(100))
    
    tags = Column(String(500))
    
    station = relationship('HelperTemplateIDs', backref='station', lazy='dynamic')
    
    @classmethod
    def fromdictionary(cls, d):
        allowed = ('name', 'mobile', 'location', 'latitude', 'longitude', 'region', 'license', 'url', 'tags')
        df = {k: v for k, v in d.items() if k in allowed}
        return cls(**df)
    
    def update(self, newdata_in_dict):
        allowed = ('name', 'mobile', 'location', 'latitude', 'longitude', 'region', 'license', 'url', 'tags')
        for key, value in newdata_in_dict.items():
            if key in allowed:
                setattr(self, key, value)
    
    def __init__(self, name="Test Station", mobile=False, location="", latitude=None, longitude=None, region=None,
                 license=None, url=None, tags=json.dumps({})):
        """
        :param name:
        :param mobile:
        :param latitude:
        :param longitude:
        :param tags:
        """
        self.name = name
        self.mobile = mobile
        self.location = location
        self.latitude = latitude
        self.longitude = longitude
        self.region = region
        self.license = license
        self.url = url
        
        self.tags = json.dumps(tags)
    
    def __eq__(self, other):
        """Override the default Equals behavior"""
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented
    
    def __ne__(self, other):
        """Define a non-equality test"""
        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented
    
    def __repr__(self):
        return '<Name %r>' % (self.name)


class AbstractObservables(Base):
    __tablename__ = "AbstractObservables"
    id = Column(Integer, primary_key=True)
    name = Column(String(60))
    ontology = Column(String(160))
    
    obs1 = relationship('HelperTemplateIDs', backref='observable', lazy='dynamic')
    
    @classmethod
    def fromdictionary(cls, d):
        allowed = ('name', 'ontology')
        df = {k: v for k, v in d.items() if k in allowed}
        return cls(**df)
    
    def __init__(self, name="Temperature", ontology=None):
        self.name = name
        self.ontology = ontology


class UnitsOfMeasurement(Base):
    __tablename__ = "UnitsOfMeasurement"
    id = Column(Integer, primary_key=True)
    name = Column(String(60))
    ontology = Column(String(160))
    symbol = Column(String(15))
    
    unit_2 = relationship('Sensors', backref='unit', lazy='dynamic')
    unit1 = relationship('HelperTemplateIDs', backref='uom', lazy='dynamic')
    
    @classmethod
    def fromdictionary(cls, d):
        allowed = ('name', 'ontology', 'symbol')
        df = {k: v for k, v in d.items() if k in allowed}
        return cls(**df)
    
    def __init__(self, name=None, ontology=None, symbol=None):
        self.name = name
        self.ontology = ontology
        self.symbol = symbol


class Sensors(Base):
    __tablename__ = "Sensors"
    """

    """
    id = Column(Integer, primary_key=True)
    generic = Column(Boolean)
    name = Column(String(60))
    manufacturer = Column(String(60))
    tags = Column(String(500))
    
    abstract_observable_id = Column(Integer, ForeignKey('AbstractObservables.id'))
    # We may need to omit the following. It does not make any sense to me.
    unit_id = Column(Integer, ForeignKey('UnitsOfMeasurement.id'))
    
    sens1 = relationship('HelperTemplateIDs', backref='sensor', lazy='dynamic')
    
    @classmethod
    def fromdictionary(cls, d):
        allowed = ('generic', 'name', 'manufacturer', 'tags', 'abstract_observable_id', 'unit_id')
        df = {k: v for k, v in d.items() if k in allowed}
        return cls(**df)
    
    def update(self, new_data_in_dict):
        allowed = ('generic', 'name', 'manufacturer', 'tags', 'abstract_observable_id', 'unit_id')
        for key, value in new_data_in_dict.items():
            if key in allowed:
                setattr(self, key, value)
    
    def __init__(self, generic=False, name="Test sensor", manufacturer="Test manufacturer",
                 abstract_observable_id=None, unit_id=None,
                 tags=json.dumps({})):
        self.generic = generic
        self.name = name
        self.manufacturer = manufacturer
        self.abstract_observable_id = abstract_observable_id
        self.unit_id = unit_id
        self.tags = json.dumps(tags)
    
    def __repr__(self):
        return '<id %r>' % (self.id)


class Observations(Base):
    __tablename__ = "Observations"
    """
    
    """
    id = Column(Integer, primary_key=True)
    timestamp = Column(DateTime)
    value = Column(String(60))
    helper_observable_id = Column(Integer, ForeignKey('HelperTemplateIDs.id'))
    
    @classmethod
    def fromdictionary(cls, d):
        allowed = ('timestamp', 'value', 'helper_observable_id')
        df = {k: v for k, v in d.items() if k in allowed}
        return cls(**df)
    
    def __init__(self, timestamp=None, value=None, helper_observable_id=None):
        """
        :param timestamp:
        :param value:
        :param flag:
        :param tags:
        """
        self.timestamp = timestamp
        self.value = value
        self.helper_observable_id = helper_observable_id
    
    def __repr__(self):
        return '<id %r>' % (self.id)


class HelperTemplateIDs(Base):
    __tablename__ = "HelperTemplateIDs"
    """

    """
    id = Column(Integer, primary_key=True)
    observable_id = Column(String(30))
    
    abstract_observable_id = Column(Integer, ForeignKey('AbstractObservables.id'))
    unit_id = Column(Integer, ForeignKey('UnitsOfMeasurement.id'))
    station_id = Column(Integer, ForeignKey('Station.id'))
    sensor_id = Column(Integer, ForeignKey('Sensors.id'))
    
    start_date = Column(DateTime)
    end_date = Column(DateTime)
    number_of_observations = Column(Integer)
    frequency = Column(String(10))
    
    helper = relationship('Station', backref="helper", cascade="all", single_parent=True)
    helper_observable_id = relationship('Observations', backref='helper', lazy='dynamic')
    
    @classmethod
    def fromdictionary(cls, d):
        allowed = ('observable_id', 'sensor_id', 'station_id', 'abstract_observable_id', 'unit_id')
        df = {k: v for k, v in d.items() if k in allowed}
        return cls(**df)
    
    def update(self, new_data_in_dict):
        allowed = ('observable_id', 'sensor_id', 'station_id', 'abstract_observable_id', 'unit_id')
        for key, value in new_data_in_dict.items():
            if key in allowed:
                setattr(self, key, value)
    
    def update_meta(self, metadata_in_dict):
        # TODO: If we append values, we have to change number of observations
        allowed = ('start_date', 'end_date', 'frequency', 'number_of_observations')
        start_date_value = getattr(self, 'start_date')
        observations = getattr(self, 'number_of_observations')
        for key, value in metadata_in_dict.items():
            if key in allowed:
                
                if key == 'start_date' and start_date_value:
                    pass
                elif key == 'number_of_observations' and observations:
                    setattr(self, key, int(value) + observations)
                else:
                    setattr(self, key, value)
    
    def __init__(self, observable_id=None, sensor_id=None,
                 abstract_observable_id=None, unit_id=None,
                 station_id=None):
        self.observable_id = observable_id
        self.sensor_id = sensor_id
        self.abstract_observable_id = abstract_observable_id
        self.unit_id = unit_id
        self.station_id = station_id
        
        self.start_date = None
        self.end_date = None
        self.number_of_observations = None
        self.frequency = None
    
    def __repr__(self):
        return f'<id {self.id!r}>'


class Templates(Base):
    __tablename__ = "Templates"
    id = Column(Integer, primary_key=True)
    filename = Column(String(60))
    path = Column(String(360))
    template_type = Column(String(20))
    for_loop_vars = Column(String(400))
    
    @classmethod
    def fromdictionary(cls, d):
        allowed = ('path', 'filename', 'template_type', 'for_loop_vars')
        df = {k: v for k, v in d.items() if k in allowed}
        return cls(**df)
    
    def __init__(self, path=None, filename=None, template_type=None, for_loop_vars=None):
        self.path = path
        self.filename = filename
        self.template_type = template_type
        self.for_loop_vars = for_loop_vars



from flask import render_template
from flask import request, make_response, jsonify
from flask_googlemaps import GoogleMaps
from flask_googlemaps import Map

from edam.viewer.app import app, calculate_metastations, nocache, calculate_templates, \
    calculate_data_and_render_from_template
from edam.viewer.app.InvalidUsage import InvalidUsage
from edam.viewer.app.OGC_SOS import OGC_SOS

GoogleMaps(app)


@app.route('/maps/')
@app.route('/index/')
@app.route('/')
def mapview():
    # creating a map in the view
    metastations = calculate_metastations()
    markers = list()
    for station_id in metastations:
        station_dict = metastations[station_id]
        if station_dict['latitude'] is not None and station_dict['longitude'] is not None:
            observables = ', '.join(
                [station_dict['observables'][temp_id]['observable'] for temp_id in station_dict['observables']])
            
            infobox = '<p><b>Station</b>: <a href="%sStations/%s" target="_blank">%s</a></p>' \
                      '<p><b>Quantities</b>: %s</p>' % (
                          request.url_root, station_id, station_dict['name'], observables)
            temp_dict = {'lat': station_dict['latitude'], 'lng': station_dict['longitude'], 'infobox': infobox}
            markers.append(temp_dict)
    mymap = Map(
        identifier="map-canvas",
        lat=38.045362,
        lng=23.715078,
        zoom=3,
        markers=markers,
        style="width:100%;height:650px;"
    )
    return render_template('maps.html', mymap=mymap)


@app.errorhandler(InvalidUsage)
def handle_invalid_usage(error):
    return render_template('error.html', errorTitle=str(error.status_code), errorMessage=error.message)


@app.route('/home/')
def index():
    return render_template('home.html')


@app.route('/Stations/')
def stations():
    metastations = calculate_metastations()
    return jsonify(metastations)


@app.route('/Stations/<station>')
def specific_station(station):
    metastations = calculate_metastations()
    return jsonify(metastations[int(station)])


@app.route('/templates/')
def templates():
    metatemplates = calculate_templates()
    return jsonify(metatemplates)


@app.route('/templates/<template>')
def specific_template(template):
    metatemplates = calculate_templates()
    try:
        path = metatemplates[template]['path']
        
        with open(path, 'r') as f:
            response = make_response(f.read())
            response.headers['Content-type'] = 'text/plain'
            return response
    except KeyError:
        raise InvalidUsage(f'{template} template does not exist', status_code=410)


@app.route('/data/')
@nocache
def get_data():
    station_id = request.args.get('station')
    observable_id = request.args.get('template')
    
    valid, template, station, chunk = calculate_data_and_render_from_template(observable_id=observable_id,
                                                                              station_id=station_id)
    # Template_id: Yucheng. The actual file is edam/Yucheng.tmpl
    if valid:
        response = make_response(render_template("edam/" + observable_id + '.tmpl', chunk=chunk, station=station))
        response.headers['Content-type'] = 'text/plain'
        response.headers['Cache-Control'] = 'public, max-age=0'
        
        # response.headers['Content-Disposition'] = "attachment; filename=test.csv"
        return response
    else:
        if station:
            raise InvalidUsage('%s template and %s station are not compatible' % (observable_id, station.name),
                           status_code=410)
        else:
            raise InvalidUsage('No station', status_code=410)


@app.route('/about/')
def about():
    return render_template('about.html')


@app.route('/SensorObservationService/')
def sensor_info():
    return render_template('SOS_Examples.html')


@app.route('/sos/')
@app.route('/Sos/')
@app.route('/sos/')
@app.route('/SOS/')
def sos():
    sos_request = request.args.get('request', '')
    sos_procedure = request.args.get('procedure', '')
    sos_offering = request.args.get('offering', '')
    sos_observed_property = request.args.get('observedProperty', '')
    sos_event_time = request.args.get('eventTime', '')
    
    sos = OGC_SOS(sos_request, sos_procedure, sos_offering, sos_event_time, sos_observed_property)
    sos_response = sos
    
    response = make_response(render_template(sos_response.template
                                             , info=sos.info
                                             , keywords=sos.keywords
                                             , stations=sos.stations
                                             , helpers=sos.metadata
                                             , procedure=sos.procedure
                                             , sensor=sos.sensor
                                             , results=sos.results
                                             , helper=sos.helper_object))
    response.headers["Content-Type"] = "application/xml"
    
    return response


if __name__ == "__main__":
    pass
    # app.jinja_env.auto_reload = True
    # app.config['TEMPLATES_AUTO_RELOAD'] = True
    # app.run(debug=True)



import logging

from edam.reader.Preprocess import Preprocess
from edam.reader.SourceConfiguration import SourceConfiguration
from edam.reader.TemplateReader import TemplateReader
from edam.reader.utilities import determine_if_file_is_csv_or_not, StorageType


class Workflow(object):
    def __init__(self, input_list, template_file=None, configuration_file=None, storage=StorageType.FILE):
        self.logger = logging.getLogger('edam.reader.Workflow.Workflow')
        
        self.template_file = template_file
        self.configuration_file = configuration_file
        
        # I instantiate this to use it later.
        
        # input_file, template_file = find_input_template_from_config_file(config_file=self.configuration_file)
        for input_file in input_list:
            self.input_file = input_file
            # set_template_table()
            # firstly we are going to preprocess input and template file,
            # To identify if input doc and template have preamble
            
            # After above call input_file and template_file names will be updated
            # More specifically, input_file will become input_file_data (containing only data)
            # The same applies to the template_file (template_file -> template_file_data)
            
            # If input has preamble, two new files will be created:
            # input_file_preamble and template_file_preamble
            
            preprocess = Preprocess(input_file=self.input_file, template_file=self.template_file)
            # This MUST have a value
            self.input_data_file = preprocess.new_input_file_data
            self.template_data_file = preprocess.new_template_file_data
            # The followings could be None
            self.input_preamble_file = preprocess.new_input_file_preamble
            self.template_preamble_file = preprocess.new_template_file_preamble
            
            # We now have to determine if the _data document is in csv format
            # If not, we have to transform it to csv format (it was obvious, wasn't it?)
            # If it's not, the following method will try and convert it in csv
            # And overwrite the content in the same file name.
            self.input_data_file, self.template_data_file = determine_if_file_is_csv_or_not(self.input_data_file,
                                                                                            self.template_data_file)
            
            # Above call returns nothing, but it creates _data and _preamble (if applicable)
            # document files
            
            # In the configuration file, a user can define ONLY ONE STATION. This is a design choice which converges
            # with the most datasets we found and tested. In addition to the the previous,
            # in a configuration file a user
            # can define MORE THAN ONE observables (a station usually has a number of them),
            # MORE THAN ONE units of measurement
            # each observable should have one (or not, e.g. we have 3 observables which have the same uom). Also, a user
            # can define MORE THAN ONE sensors, usually each one for an observable.
            # However, there are cases wherein the exact
            # metadata about specific sensors are not known. In these cases, a user can create ONE GENERIC SENSOR and
            # define the required relationships (via 'relevant_observables' value).
            
            # What about the cases (eg. KNMI) a file contains an aggregated view of all available stations?
            # All stations observe the exact same observables,
            # which are observed from the very same sensors (not in the same
            # location, though). And all observables are represented with the very same uoms.
            config = SourceConfiguration(input_yaml=self.configuration_file,
                                         input_file_data=self.input_data_file,
                                         template_preamble=self.template_preamble_file,
                                         input_preamble=self.input_preamble_file)
            
            # Important questions.
            # 1. What to expect in the preamble?
            # 2. How are we going to represent multiple stations? Consider the knmi data input. Let's somehow assume
            # that we parsed stations in the preamble. When are we going to create the helperTable??
            # This table is a quick and cheap way to keep track of observations,
            # so as each observations is self-describable.
            #
            # We are now ready to call the template reader class.
            # try:
            TemplateReader(config=config, input_file=self.input_data_file, template=self.template_data_file)
            # except Exception as e:
            #     self.logger.error("Exception: %s %s" % (type(e).__name__, str(e)))
            self.input_data_file.close()
            self.input_file.close()


if __name__ == "__main__":
    pass
    # t = Workflow()



import copy
import io
import logging
import re
from datetime import datetime

import numpy as np
import pandas as pd

from edam.reader.SourceConfiguration import SourceConfiguration
from edam.reader.manage import DatabaseHandler
from edam.reader.models import HelperTemplateIDs

var_name = re.compile(r"{{(.*?)}}")
var_for_loop = re.compile(r"{%\s?for (.*?) in (.*?)%}")
var_end_for_loop = re.compile(r"{%endfor%}")

# Above regex is for the case of same_timestamp(function) found in BoM template
var_same_timestamp = re.compile(r"(same_timestamp)\s?\((?P<args>(?P<arg>[\w.]+(,\s?)?)+)\)")

# This var corresponds to observables which value is located in more than one columns
# In our templating language they are represented as: {{observable.value[0]}},{{observable.value[1]}}, etc.
# With that regex we are going to determine the existence of those observables
var_multicolumn_values = re.compile(r"{{.*?\..*?(\[[0-9]\])*}}")

# For the following regex, line has to be splitted to ',' and then
# each individual var to be parsed with var_name
var_for_line = re.compile(r"{%\s?for .*? in .*?%}\n(.*)\n{%\s?endfor\s?%}")
var_parse_header = re.compile(r"(.*?)\n*{%\s?for .*? in .*?%}")

# var_anything_but_value = re.compile(r"{{.*?.value(.*?)}}")
# var_anything_but_value = re.compile(r"{{.+.value(\[.+\]).?.?}}")
var_for_value_timestamp = re.compile(r"{{.+.(timestamp.?)}}")

module_logger = logging.getLogger('edam.reader.TemplateReader')


def __determine_how_to_group_observables__(df_columns_indexed: dict):
    """
    This function determines how to parse observables. Some observables may have values and/or timestamps which
    expand in more than one columns. That function determines that and returns a list with indices.
    e.g. [1, 2, 3, 4, [5, 6, 7], 8, 9, [10, 11]]
    Above list means that 1, 2, 3, 4, 8, 9 should be parsed independently. 5,6,7 and 10,11 should be parsed
    as groups.
    In the returned list, if variable is "value" then index is of int type. If it's timestamp is of str type.
    E.g. In the list: [0, 1, 2, 3, 4, [5, 6, '7'], 8, 9, 10, [11, 12], 13, 14, 15, 16, [17, 18], 19]
    
    7 column contains timestamp.
    #TODO determine the type of timestamp.
    :param df_columns_indexed: Dictionary (index: value) (e.g. {0: '{{tmax.value}}', 1: '{{tmin.value}}', 2:
    '{{af.value}}', 3: '{{rain.value}}', 4: '{{sun.value}}'})

    :return:
    """
    
    how_to_parse = []
    temp = dict()
    args = list()
    for ind, var in df_columns_indexed.items():
        match_timestamp = re.search(var_for_value_timestamp, var)
        # match_multi_value = re.search(var_anything_but_value, var)
        index = copy.deepcopy(ind)
        # if match_timestamp or match_multi_value:
        argument = re.findall(var_name, var)[0]
        if match_timestamp:
        
            if "timestamp" in argument:
                index = str(index)
            
            argument = argument.strip(' \n\r').split('.')
            if len(argument) > 2:
                argument = argument[0] + argument[-1]
            else:
                argument = argument[0]
            
            args.append(argument)
            if argument not in temp.keys():
                temp[argument] = list()
            temp[argument].append(index)
        else:
            argument = argument.strip(' \n\r').split('.')[0]
            if argument in temp.keys():
                temp[argument].append(index)
            else:
                search_for = argument+".timestamp"
                if [s for s in df_columns_indexed.values() if search_for in s]:
                    temp[argument] = list()
                    temp[argument].append(index)
                else:
                    how_to_parse.append(index)
    while bool(temp):
        how_to_parse.append(temp.popitem()[1])

    return how_to_parse


def __get_statements_from_placeholders__(placeholder: str):
    return placeholder.split('=')


class TemplateReader:
    """
    This class takes 2 inputs:
    1. Source configuration object
    2. Connector object

    From those it infers:
    1. template
    2. input_file

    and exports:
    3. A list with observation objects, ready to be stored in the database.
    """
    
    def __init__(self, config=None, input_file=io.StringIO(),
                 template=io.StringIO()):
        self.template_logger = logging.getLogger('edam.reader.TemplateReader.TemplateReader')
        self.input_file = input_file
        self.template = template
        
        self.config = config
        
        self.Data = DatabaseHandler()
        
        self.df = None
        self.same_timestamp_arguments = None
        
        # I will create tuples (station,respective_dataframe,for_lines_indexed)
        # and append them in a list (parsing_tuples)
        # This process will be the first step. Parsing/storing and further processing follows.
        # station: The database id of the station for which data will be parsed
        # IMPORTANT: In case the input data is row-based and not column based
        # (i.e. australian data, see also git issue 6), we will generate a dataframe which will contain as many columns
        # as the "unique" observables of the station.
        
        # respective_dataframe: A dataframe which will have as index the timestamp column, and its related observables
        # will be located on the other df columns. Parsing/storing of such a df is already implemented.
        
        # for_lines_indexed: each dataframe should have its own for_line_indexed dictionary. Consider the example
        # of australian data. We have a number of station, each of which CAN POTENTIALLY have different observables..
        
        self.parsing_tuples = list()
        
        self.__open_template__()
        self.__set_dataframe_index_col()
        
        self.__create_dataframe_from_csv__()
        self.template_logger.info("I created the df from the csv")
        
        self.template_logger.info("I am starting handling stations")
        
        self.__handle_station_column__()
        self.template_logger.info("I am parsing data now")
        
        for station_id, station_respective_df, for_lines_indexed in self.parsing_tuples:
            self.template_logger.info("I am parsing station with %s id" % station_id)
            rows, columns = station_respective_df.shape
            self.template_logger.info("Rows: %d, Columns: %d" % (rows, columns))
            how_to_parse = __determine_how_to_group_observables__(df_columns_indexed=for_lines_indexed)
            self.__generate_pandas_series_from_df__(station_dataframe=station_respective_df,
                                                    how_to_parse=how_to_parse,
                                                    df_columns_indexed=for_lines_indexed,
                                                    station_id=station_id)
    
    def __open_template__(self):
        """
        This function open the self.template file and stores header (self.template_header_indexed) as dictionary
        and for_lines_arguments (self.template_for_lines_indexed) as dictionary.
        :return:
        """
        self.template.seek(0)
        text = self.template.read()
        # parse header
        header = re.findall(var_parse_header, text)[0].strip('\r\n').split(',')
        # create a dictionary with indices
        header = dict(enumerate(header))
        # TODO: Please check if the following is needed and remove.
        # self.template_header_indexed = dict(enumerate(header))
        
        # TODO: Please check if the following is needed and remove.
        for_lines = re.findall(var_for_line, text)[0].strip('\r\n').split(',{{')
        
        # for_lines = re.findall(var_for_line, text)[0].strip('\r\n')
        # for_lines = re.findall(var_name, for_lines)

        for_lines = list(map(lambda x: x if x.startswith("{{") else "{{" + x, for_lines))

        self.usecols = dict(enumerate(for_lines))
        self.usecols = list(filter(lambda key: re.search(var_name, self.usecols[key]), self.usecols))
        
        # Parse only values that need to be parsed (i.e. those inside placeholders{{}})
        for_lines = list(filter(lambda x: re.search(var_name, x), for_lines))
        
        self.template_for_lines_indexed = dict(enumerate(for_lines))
        self.for_lines = list(filter(lambda x: re.search(var_name, x), for_lines))
        self.for_lines = list(map(lambda x: x.strip("{}\r\n"), self.for_lines))
        self.template_header = dict()
        for index, label in self.template_for_lines_indexed.items():
            label_to_be_stored = label.strip("{}")
            if label_to_be_stored in self.template_header.keys():
                # TODO! This is so static, just to work for BoM! EDIT
                label_to_be_stored = label_to_be_stored + ".1"
            try:
                self.template_header[label_to_be_stored] = header[index]
            except Exception as e:
                self.template_logger.error(
                    "Can't create self.template_header for %s Exception: %s %s" % (
                        label_to_be_stored, type(e).__name__, str(e)))
        self.template.seek(0)
        
    def __set_dataframe_index_col(self):
        """
        This function creates a list with all values that will be passed over (self.no_parse_vars).
        It looks all for_line_vars, finds the one starts with "timestamp" and sets that column as the index of the
        dataframe.
        TODO: Make this more generic
        :return:
        """
        # In some cases timestamp could be extended in more than one columns
        # E.g. 09.05.2014,14:23:34,0.004
        # Thus index should be a dictionary: {date: index1, month: index2, time:index3}
        # Columns representing above indices should be merged when construction of dataframe takes place.
        self.parse_dates = dict()
        self.parse_dates['timestamp'] = dict()
        self.parse_dates['timestamp']['indices'] = list()
        self.parse_dates['timestamp']['format'] = list()
        self.index = {}
        for index, variable in self.template_for_lines_indexed.items():
            match = re.search(var_name, variable)
            match_same_timestamp = re.search(var_same_timestamp, variable)
            
            if match:
                if match_same_timestamp:
                    match_same_timestamp = re.search(var_same_timestamp, variable)
        
                    fn_dict = match_same_timestamp.groupdict()
                    arguments = [arg.strip() for arg in fn_dict['args'].split(',')]
                    # We pick one of the two, since they are the same
                    # argumemnts = ['windm_spd.timestamp.time', 'windm_dir.timestamp.time']
                    self.same_timestamp_arguments = arguments
                    name_of_variable_without_brackets = arguments[0]
                else:
                    name_of_variable_without_brackets = re.findall(var_name, variable)[0]
                    # name_of_variable_without_brackets: timestamp.date
                    # Thus splitting by '.' and taking the last item, ie. date, time, etc.
                    dict_key = name_of_variable_without_brackets.split('.')[-1]
                if name_of_variable_without_brackets.startswith("timestamp"):
                    self.parse_dates['timestamp']['indices'].append(index)
                    if dict_key.lower() == "year":
                        self.parse_dates['timestamp']['format'].append('%Y')
                    elif dict_key.lower() == "month":
                        self.parse_dates['timestamp']['format'].append('%m')
                    elif dict_key.lower() == "day":
                        self.parse_dates['timestamp']['format'].append('%d')
                    elif dict_key.lower() == "dayofyear":
                        self.parse_dates['timestamp']['format'].append('%j')
                    elif dict_key.lower() == "hour":
                        self.parse_dates['timestamp']['format'].append('%H')
                    elif dict_key.lower() == "minutes":
                        self.parse_dates['timestamp']['format'].append('%M')
                    elif dict_key.lower() == "seconds":
                        self.parse_dates['timestamp']['format'].append('%S')
                    else:
                        self.template_logger.debug("%s timestamp type" % dict_key)
                        pass
                elif "timestamp" in name_of_variable_without_brackets:
                    # all cases where timestamp is not first (e.g. wind.timestamp..)
                    additional_timestamp = name_of_variable_without_brackets.split('.')[0] + '.timestamp'
                    if additional_timestamp not in self.parse_dates:
                        self.parse_dates[additional_timestamp] = dict()
                        # All sub-timestamps depend on the main timestamp
                        self.parse_dates[additional_timestamp]['indices'] = copy.deepcopy(
                            self.parse_dates['timestamp']['indices'])
                    
                    self.parse_dates[additional_timestamp]['indices'].append(index)
    
    def __create_dataframe_from_csv__(self):
        """
        At the end of this function, df has as index the correct timestamp and has all non-relevant columns dropped.
        We still need to parse static information (such as timestamps) from the header, ie. the column names.
        :return:
        """
        # I don't parse the header of the input file. Instead I set the header to the template to be used for the
        # names of the df columns.
        if self.df is None:
            parse_dates = dict()
            for key, value in self.parse_dates.items():
                parse_dates[key] = value['indices']
            if self.parse_dates['timestamp']['format']:
                def date_parser(x):
                    try:
                        return datetime.strptime(x, ' '.join(self.parse_dates['timestamp']['format']))
                    except:
                        # This exception catches the case where in datetime column we have litter (e.g. Site closed)
                        return x
            else:
                date_parser = None
            self.df = pd.read_csv(self.input_file, na_values='---',
                                  header=0,
                                  usecols=self.usecols,
                                  names=self.for_lines,
                                  warn_bad_lines=True,
                                  parse_dates=parse_dates,
                                  date_parser=date_parser,
                                  infer_datetime_format=True,
                                  keep_date_col=False,
                                  error_bad_lines=False,)
            self.df.set_index(keys=["timestamp"], drop=True, inplace=True)
            # Drop nan rows and columns
            self.df.dropna(axis=0, how='all', inplace=True)
            self.df.dropna(axis=1, how='all', inplace=True)
            # Let's create duplicate lines of same_timestamp_arguments (if any)
            if self.same_timestamp_arguments:
                arguments = copy.deepcopy(self.same_timestamp_arguments)
                arguments = list(map(lambda x: x.split('.')[0] + '.timestamp', arguments))
                key_argument = copy.deepcopy(arguments[0])
                del arguments[0]
                # The following if clause is for bom data. When no windmax values are given...
                try:
                    check_list = list(filter(lambda x: 'nan' not in x, self.df[key_argument].values.tolist()))
                except:
                    check_list = True
                    
                if check_list:
                    for arg in arguments:
                        self.df[arg] = self.df[key_argument]
                else:
                    self.df.drop(key_argument, axis=1, inplace=True)
                
        else:
            pass

    def __handle_station_column__(self):
        temp_parsing_tuples = list()
        for key in list(self.template_for_lines_indexed):
            column_name = self.template_for_lines_indexed[key].strip('{{}}')
            
            # I assume that if station is mentioned in the observations iteration, it will be the FK to the station
            # to which the datapoints are referring to.
            # TODO: However, this should be more generic and predict any other unforeseen cases.
            if 'station' in column_name:
                distinct_stations = self.df[column_name].unique()
                
                for tags_station_id in distinct_stations:
                    # If the station is in the ones the users defined in their config file...
                    database_station_id = self.__return_dbstation_id_by_tags_value__(tags_station_id)
                    if database_station_id in self.config.station_id:
                        respective_station_df = self.df.loc[self.df[column_name] == tags_station_id].copy()
                        respective_station_df.drop(column_name, axis=1, inplace=True)
                        temp_parsing_tuples.append((database_station_id, respective_station_df))
                    else:
                        self.template_logger.warning("You are trying to parse station with database id '%s'. \n"
                                                     "However your input file does not contain "
                                                     "a station with this id. \n Program will exit shortly... \n" %
                                                     tags_station_id)
                
                self.template_for_lines_indexed.pop(key)
                break
        
        # Update template_for_line_indexed dictionary
        # This is an essential step since data parsing is based on this dictionary
        # Now let's check if we have the case of the australian data.
        # That is, if we have row-based data
        
        # if 'observable.observable_id' == column_name:
        #     unique_observables = self.df[column_name].unique()
        
        # I will now create a new df column for each of those unique observables.
        # The values of those
        
        temp_temp_parsing_tuples = list()
        
        if '{{observable.observable_id}}' in self.template_for_lines_indexed.values():
            for station, resp_df in temp_parsing_tuples:
                resp_df = resp_df.pivot_table(index=resp_df.index.name, columns='observable.observable_id',
                                              values='observable.observable_id.value')
                # resp_df = resp_df.pivot(columns='observable.observable_id', values='observable.observable_id.value')
                # try:
                #     resp_df = resp_df.pivot(columns='observable.observable_id', values='observable.observable_id.value')
                # except:
                #     print(resp_df)
                temp_temp_parsing_tuples.append((station, resp_df))
            del temp_parsing_tuples
            temp_parsing_tuples = temp_temp_parsing_tuples
        del temp_temp_parsing_tuples
        # If list is empty!
        if not temp_parsing_tuples:
            # I assume that in config one and only one station was defined..
            one_tuple = (self.config.station_id[0], self.df)
            temp_parsing_tuples.append(one_tuple)
            # self.parsing_tuples.append(one_tuple)
        
        # At this point we have dataframes which have timestamp as index, and other columns are for the observables
        for station_id, station_df in temp_parsing_tuples:
            temp_for_lines_indexed = dict()
            for column_name in list(station_df):
                # TODO: curly brackets are used for continuity purposes
                temp_for_lines_indexed[station_df.columns.get_loc(column_name)] = "{{" + column_name + "}}"
            tuple_to_be_added = station_id, station_df, temp_for_lines_indexed
            self.parsing_tuples.append(tuple_to_be_added)
    
    def __update_index_timestamp_from_column__(self, station_df, var, grouped_col=False):
        # Do something with only columns with additional timestamps
        # check if we need to parse data from header
        regex = re.compile(r'{%.?set (.*?).?%}')
        try:
            local_var = self.template_header[var]
        except Exception as e:
            self.template_logger.warning("%s %s" % (type(e).__name__, str(e)))
            self.template_logger.warning("I am returning df without updates")
            new_df = pd.DataFrame()
            new_df["value"] = pd.Series(station_df[var], index=station_df.index)
            return new_df
        match = re.search(regex, local_var)
        if grouped_col:
            new_df = station_df
        
        else:
            new_df = pd.DataFrame()
            new_df["value"] = pd.Series(station_df[var], index=station_df.index)
        if match:
            # example: {{wind.timestamp.hour=9}}
            fullname_in_list = re.findall(regex, local_var)[0]
            
            # example: ['wind.timestamp.hour', '9']
            splitted_by_equal_sign = __get_statements_from_placeholders__(fullname_in_list)
            
            # example: hour
            # but we need only the first letter
            # thus [0]
            unit = splitted_by_equal_sign[0].split('.')[-1][0]
            
            new_df.index += pd.TimedeltaIndex(pd.Series(np.full(new_df.shape[0],
                                                                int(splitted_by_equal_sign[1]))), unit=unit)
        
        return new_df
    
    def __determine_observable_id_from_db__(self, var, station_id) -> HelperTemplateIDs:
        """
        :param var: {{temp.value}}
        :return:
        """
        var = re.findall(var_name, var)[0].split('.')[0]
        helper_template_row = self.Data.__get_helper_table_row_input_file_observable_id__(var,
                                                                                        station_id)
        
        return helper_template_row
    
    def __generate_pandas_series_from_df__(self, station_dataframe, how_to_parse, df_columns_indexed, station_id):
        for col_index in how_to_parse:
            # if True it means the column should be parsed independently.
            # if False (list type), another for loop should be implemented
            # dataframe_to_store = pd.DataFrame()
            if type(col_index) is int:
                col_index = int(col_index)
                var = station_dataframe[station_dataframe.columns[col_index]].name
                dataframe_to_store = self.__update_index_timestamp_from_column__(station_df=station_dataframe,
                                                                                 var=var)  # type: pd.DataFrame()
                helper_id_row = self.__determine_observable_id_from_db__(df_columns_indexed[col_index], station_id)
                
                dataframe_to_store["helper_observable_id"] = pd.Series(helper_id_row.id, index=dataframe_to_store.index)
                
                update_helper_with_meta = dict()
                try:
                    update_helper_with_meta['frequency'] = pd.infer_freq(dataframe_to_store.index)
                    update_helper_with_meta['start_date'] = dataframe_to_store.index[0]
                    update_helper_with_meta['end_date'] = dataframe_to_store.index[-1]
                    update_helper_with_meta['number_of_observations'] = dataframe_to_store.__len__()
                    self.__update_helper_observable_id(helper_id_row, update_helper_with_meta)
                except:
                    pass
            else:
                dataframe_to_store = pd.DataFrame(index=station_dataframe.index)
                exits = False
                for grouped_column in col_index[:]:
                    if type(grouped_column) is str:
                        grouped_column = int(grouped_column)
                        
                        col_index.remove(str(grouped_column))
                        exits = True
                        break
                dataframe_to_store['value'] = station_dataframe.iloc[:, col_index].astype(str).apply(
                    lambda x: ' '.join(x),
                    axis=1)
                if exits and type(grouped_column) is int:
                    dataframe_to_store.index = station_dataframe[
                        station_dataframe[station_dataframe.columns[grouped_column]].name]
                
                # We don't care which col_index we will select. They are both referring to the same entity
                # They are grouped after all
                var = station_dataframe[station_dataframe.columns[col_index[0]]].name
                # Update index from column
                
                dataframe_to_store = self.__update_index_timestamp_from_column__(
                    station_df=dataframe_to_store, var=var, grouped_col=True
                )  # type: pd.DataFrame()
                
                helper_id_row = self.__determine_observable_id_from_db__(
                    df_columns_indexed[int(col_index[0])], station_id)
                dataframe_to_store["helper_observable_id"] = pd.Series(helper_id_row.id,
                                                                     index=dataframe_to_store.index)
                update_helper_with_meta = dict()
                try:
                    update_helper_with_meta['frequency'] = pd.infer_freq(dataframe_to_store.index)
                    update_helper_with_meta['start_date'] = dataframe_to_store.index[0]
                    update_helper_with_meta['end_date'] = dataframe_to_store.index[-1]
                    update_helper_with_meta['number_of_observations'] = dataframe_to_store.__len__()
                    self.__update_helper_observable_id(helper_id_row, update_helper_with_meta)
                except:
                    pass
            
            # Clean from duplicate records
            # dataframe_to_store = self.Data.clean_df_db_dups(df=dataframe_to_store, tablename="Observations",
            #                                                 dup_cols=list(dataframe_to_store))

            self.Data.__add_dataframe__(dataframe_to_store)
    
    def __check_if_observable_is_stored__(self, observable):
        return self.Data.__check_observable_is_in_db__(observable)
    
    def __return_dbstation_id_by_tags_value__(self, station_id):
        """
        #TODO: This function should be more generic. E.g. By placeholder value
        In the template we have something like: {{station.tags.station_id}}
        Logic of this program is "smart" enough to identify that station_id is a key of the JSON type "tag".
        Thus, this can be handled automatically in the future..
        :param station_id: tags station_id, e.g. 210
        :return: a dictionary. old_value:new_value
        """
        
        database_station_id = self.Data.__get_station_id_by_tags_station_id__(station_id)
        # for old_value, new_value in temp_dict.items():
        #     self.df[column_name] = self.df[column_name].replace(to_replace=old_value, value=new_value)
        return database_station_id
    
    def __check_if_sensor_is_stored__(self, sensor):
        """
        
        :param sensor:
        :return: True if it exists, False if it does not
        """
        return self.Data.__check_sensor_is_in_db__(sensor)
    
    def __store_item_in_db(self, item):
        self.Data.__add_item__(item)
    
    def __update_helper_observable_id(self, helper_observable_id: HelperTemplateIDs, meta_dictionary):
        # helper_observable_id.update_meta(metadata_in_dict=meta_dictionary)
        self.Data.__update_item__(helper_observable_id, metadata_dict=meta_dictionary)


if __name__ == "__main__":
    conf = SourceConfiguration(input_yaml="/Users/argyris/Documents/git/templateFramework/metadata/knmi.yaml",
                               input_file_data="/Users/argyris/Documents/git/templateFramework/inputs/knmi_alldata_data.txt")
    t = TemplateReader(config=conf,
                       input_file="/Users/argyris/Documents/git/templateFramework/inputs/knmi_alldata_data.txt",
                       template="/Users/argyris/Documents/git/templateFramework/templates/knmi_data.tmpl")



import copy
import csv
import errno
import fnmatch
import io
import itertools
import json
import logging
import os
import re
from datetime import datetime
from enum import Enum
import numpy as np
import pandas as pd
import records
import requests
import yaml

from edam.reader.manage import DatabaseHandler
from edam.reader.models import Station
from edam.settings import database_type, home_directory

utilities_logger = logging.getLogger('edam.reader.utilities')


class InputType(Enum):
    FILE = 1
    FOLDER = 2
    DATABASE = 3
    HTTP = 4
    

class StorageType(Enum):
    FILE = 'file'
    MEMORY = 'memory'


def find_and_describe_templates(directory=home_directory):
    temp_dict = dict()
    for fold, _, files in os.walk(os.path.join(directory, 'templates')):
        for file in files:
            filename = os.path.join(fold, file)
            if fnmatch.fnmatch(filename, '*.tmpl'):
                name, ext = os.path.splitext(file)
                temp_dict[name] = dict()
                temp_dict[name]['path'] = filename
                temp_dict[name]['variables'] = get_template_for_vars(open(filename, 'r'))
    
    return temp_dict


def find_and_describe_template(template):
    temp_dict = dict()
    
    exists, file_type, file_path, file_object = check_if_path_exists(template + ".tmpl")
    if exists and file_type is InputType.FILE:
        temp_dict['path'] = file_path
        temp_dict['variables'] = get_template_for_vars(file_object)
        
        return temp_dict
    return None


def combine64(years, months=1, days=1, weeks=None, hours=None, minutes=None,
              seconds=None, milliseconds=None, microseconds=None, nanoseconds=None):
    # Above line was added to fix the drawback when in a year column we had a string (e.g. "Site Closed")
    years = pd.to_numeric(years, errors='coerce').dropna().astype(int)
    years = np.asarray(years) - 1970
    
    try:
        months = pd.to_numeric(months, errors='coerce').dropna().astype(int)
        months = np.asarray(months) - 1
    except AttributeError:
        # It means user didn't give a month parameter!!
        months = np.asarray(months) - 1
    
    try:
        days = pd.to_numeric(days, errors='coerce').dropna().astype(int)
        days = np.asarray(days) - 1
    except AttributeError:
        # It means user didn't give a day parameter!!
        days = np.asarray(days) - 1
    types = ('<M8[Y]', '<m8[M]', '<m8[D]', '<m8[W]', '<m8[h]',
             '<m8[m]', '<m8[s]', '<m8[ms]', '<m8[us]', '<m8[ns]')
    vals = (years, months, days, weeks, hours, minutes, seconds,
            milliseconds, microseconds, nanoseconds)
    return sum(np.asarray(v, dtype=t) for t, v in zip(types, vals)
               if v is not None)


def space_delimited_to_csv(input_file: io.StringIO):
    """
    I copied that from:
    http://stackoverflow.com/questions/19759423/convert-a-space-delimited-file-to-comma-separated-values-file-in-python
    
    :param input_file:
    :return:
    """
    input_file.seek(0)
    input_text = input_file
    output_text = io.StringIO()
    o = csv.writer(output_text)
    for line in input_text:
        stripped_line = line.strip('\n')
        # Don't put commas in for loops
        if ("{%for" or "{%endfor}") in stripped_line:
            output_text.write(line)
        else:
            o.writerow(stripped_line.split())
    return output_text


def extract_data_from_preamble(station: Station, preamble_template, preamble_input):
    """
    This function extracts station information from preambles.
    Afterwards, station object is updated and is returned
    :param station: models.Station object
    :param preamble_template:
    :param preamble_input:
    :return:
    """
    # Regex to identify placeholder values (e.g. {{station.name}})
    var_name = r"({{.*?}})"
    # Returns 0 (ie False if empty)
    if preamble_template.seek(0, os.SEEK_END) > 0 and preamble_input.seek(0, os.SEEK_END) > 0:
        station_dictionary = dict()
        station_dictionary['tags'] = dict()
        preamble_template.seek(0)
        preamble_input.seek(0)
        ptemp = preamble_template
        
        pinp = preamble_input
        
        for template_line, input_line in itertools.zip_longest(ptemp, pinp):
            # Sometimes in input file we have more lines than we thought we would have
            # (when we were drafting the template file). Thus, template_file variable is None
            # and program breaks. This is why we havethe following if
            if template_line:
                matches = re.findall(var_name, template_line)
                
                if matches:
                    for match in matches:
                        
                        # input_line = 'Location: 359800E 223800N, Lat 51.911 Lon -2.584, 67 metres amsl'
                        template_line = template_line.lstrip('')
                        input_line = input_line.lstrip('')
                        # template_line = 'Location: 359800E 223800N, Lat {{station.latitude}} Lon
                        # {{station.longitude}}, {{station.tags.altitude}}'
                        
                        # new_template_line = ' Lon {{station.longitude}}, {{station.tags.altitude}}'
                        
                        # to_be_replaced = 'Location: 359800E 223800N, Lat'
                        to_be_replaced = template_line.partition(match)[0].strip('\n\r')
                        template_line = template_line.partition(match)[-1].strip('\n\r')
                        
                        # value_of_placeholder = input_line.partition(template_line.partition('{')[0])[0]
                        if to_be_replaced.strip(' ') == "":
                            value_of_placeholder = input_line
                        else:
                            value_of_placeholder = input_line.replace(to_be_replaced, '').strip('\n\r')
                        # print(value_of_placeholder)
                        temp_more_curly = template_line.partition('{')[0].strip('\n\r')
                        
                        if temp_more_curly == '':
                            pass
                        else:
                            input_line = value_of_placeholder
                            value_of_placeholder = value_of_placeholder.partition(temp_more_curly)[0]
                            input_line = input_line.replace(value_of_placeholder, '')
   
                        # eg ['station', 'latitude'] or ['station', 'tags', 'key']
                        placeholder_var_in_list = match.strip("{}").split('.')
                        # remove 'station', ie first element
                        del placeholder_var_in_list[0]
                        # Now it should be either ['latitude'] or ['tags', 'key']
                        if placeholder_var_in_list[0] == 'tags':
                            station_dictionary['tags'][placeholder_var_in_list[-1]] = value_of_placeholder
                        else:
                            station_dictionary[placeholder_var_in_list[0]] = value_of_placeholder
        
        try:
            station_dictionary['tags'] = json.dumps(station_dictionary['tags'])
        except:
            pass
        station.update(station_dictionary)
    return station


def determine_if_file_is_csv_or_not(input_data: io.StringIO, template_data):
    """
    This function is based on a dummy assumption:
    If the count of `space` characters is bigger than the  count of `comma` characters in a document
    then it is not a csv.
    Following this, it transforms it using space_delimitted_to_csv
    :param input_data: input's data as a io.StringIO()
    :param template_data: template's data as a io.StringIO()
    :return: input_data, template_data (converted as csv or remained intact)
    """
    input_data.seek(0)
    input_text = input_data.read()
    if (input_text.count(' ') > input_text.count(',')) and input_text.count(',') < 10:
        # Change them!
        input_data = space_delimited_to_csv(input_file=input_data)
        input_data.seek(0)
        template_data = space_delimited_to_csv(input_file=template_data)
        template_data.seek(0)
    input_data.seek(0)
    return input_data, template_data


def parse_for_iterations(input_iteration_file, template_iteration_file, iterable_type='Station'):
    """
    I presume that files are formatted as csv.
    The strategy here is as follows:
    1. Clear anything that is not enclosed in {{}}. E.g. ":".
    TODO: Do the same for actual observations
    2. Create a dataframe from the csv file
    3. Change dataframe column names to the information derived from template. E.g. 1st column is station.station_id
    (strip "station.". !Attention .split('.')[-1])
    4. Drop any columns we don't want to really parse
    
    :param input_iteration_file:
    :param template_iteration_file:
    :param iterable_type: 'Station', 'Observables', 'Unit of Measurements'
    :return:
    """
    var_for_line = re.compile(r"{%for .*? in .*?%}\n(.*)\n{%endfor%}")
    var_name = re.compile(r"({{.*?}})")
    
    temp_iteration_file = template_iteration_file
    temp_iteration_file.seek(0)
    text = temp_iteration_file.read()
    
    # for_lines = "#{{station.station_id}}:,{{station.longtitude}},{{station.latitude}}"
    for_lines = re.findall(var_for_line, text)[0]
    # variables = ['{{station.station_id}}', '{{station.longtitude}}', '{{station.latitude}}']
    variables = [var for var in re.findall(var_name, for_lines)]
    
    characters_to_be_replaced = for_lines
    for variable in variables:
        characters_to_be_replaced = characters_to_be_replaced.replace(variable, '')
    # characters_to_be_replaced = "#:,,,,"
    # Thus we should remove comma character
    characters_to_be_replaced = characters_to_be_replaced.replace(',', '')
    for character_to_be_replaced in characters_to_be_replaced:
        for_lines = for_lines.replace(character_to_be_replaced, '')
    for_lines = for_lines.split(',')
    template_for_lines_indexed = dict(enumerate(for_lines))
    
    # Determine which indexes hold variables
    # Create dataframe header
    dataframe_header = []
    for counter_index in range(0, len(template_for_lines_indexed)):
        stripped_header = template_for_lines_indexed[counter_index].strip("{{}}")  # type: str
        stripped_header = stripped_header.split('.')
        # Remove station
        iterative_type = stripped_header[0]
        del stripped_header[0]
        # This is done for tags.something
        stripped_header = '.'.join(stripped_header)
        dataframe_header.append(stripped_header)
    
    df = pd.read_csv(input_iteration_file, na_values='', header=0, names=dataframe_header)
    
    for column in df.columns:
        # first strip
        try:
            df[column] = df[column].str.strip(to_strip=characters_to_be_replaced)
        except:
            pass
        if column.startswith('tags'):
            key = str(column.split('.')[-1])
            temp_dict = dict()
            temp_dict[key] = df[column].map(str)
            
            # Update column, key: value
            df[column] = "\"" + key + "\"" + ':' + "\"" + df[column].map(str) + "\""
            
            if 'tags' in df.columns:
                df['tags'] = df['tags'] + ',' + df[column]
            else:
                df['tags'] = '{' + df[column].map(str)
            # Drop this column
            df.drop(column, axis=1, inplace=True)
    # Exception will occur for observables for which we don't have tags
    try:
        df['tags'] = df['tags'] + '}'
        # df['tags'] = df['tags'].apply(lambda x: json.loads(x))
    except:
        pass
    
    # Station data(metadata) are stored directly to the database, while observables are sent back to
    # SourceConfiguration for further processing
    
    # just check the first column if 'observable' keyword is used.
    if iterative_type == 'observable':
        temp_iteration_file.seek(0)
        return df.to_dict(orient='index')
    
    else:
        data = DatabaseHandler()
        # Remove tags json field. It creates problems with pd.merge
        
        dup_cols = list(df)
        try:
            dup_cols.remove('tags')
        except:
            pass
        if database_type == "postgres":
            tablename = 'public."' + iterable_type + '"'
        elif database_type == "sqlite":
            tablename = '"' + iterable_type + '"'
        df_to_store = data.clean_df_db_dups(df=df, tablename=tablename, dup_cols=dup_cols)  # type: pd.DataFrame
        
        data.__add_dataframe__(dataframe=df_to_store, table=iterable_type, index=False, index_label=None)
        temp_iteration_file.seek(0)
        return [int(x) for x in df_to_store.index.tolist()]


def get_template_file_header_columns(template_file):
    """
    We sum the observables of a template
    :param template_file: 
    :return list:
    """
    template_file_text = template_file.read()
    template_file.seek(0)
    separator = "({.*?}+)"
    argument_list = list()
    
    # Each line is a list element
    lines = template_file_text.splitlines()
    
    # Remove empty lines
    [argument_list.append(remove_template_placeholders_from_string(line, separator)) for line in lines]
    argument_list = list(filter(str.strip, argument_list))
    return argument_list


def remove_template_placeholders_from_string(string, separator="({.*?}+)"):
    """
    This function cleans a **string** according to a given **separator**.
    
    :param string: The string which will be cleaned
    :param separator: Regex pattern. $param1 will be cleaned according to this param.
    :return: string: Cleaned string
    """
    
    temp_list = re.split(separator, string, 1)
    # string does not contain separator. Return as is
    if len(temp_list) == 1:
        returning_value = temp_list[0]
    else:
        # It means it has 3 elements.
        # returning_value = temp_list[0] + remove_template_placeholders_from_string(temp_list[2], separator)
        returning_value = temp_list[0]
    
    return returning_value


def check_template_against_input_file(template_file: io.StringIO, data_input: io.StringIO):
    """
    We are going to work with every line (i.e. list element) individually.
    Specifically, we are going to check each of the lines against input files.
    If all lines match to a given input file, it means that we can parse data out of it.

    :param template_file: The name of the template (e.g. Yucheng.tmpl)
    :param data_input: Either data_input filename (e.g. Yucheng.met) or string with content stored in memory
    :param input_file_type: Either "File" or anything else (check data_input parameter)
    :return: True in case match is 100% or False in case match is NOT 100%
    """
    
    # utilities_logger.debug("Template: %s" % template_file.read())
    check_lines = get_template_file_header_columns(template_file=template_file)
    utilities_logger.debug("Check lines: %s" % check_lines)
    matching_percentage = int()
    
    utilities_logger.debug("Input file: %s" % data_input)
    
    input_lines = ''.join(list(itertools.islice(data_input, len(check_lines)))).strip('\n')
    
    for line in check_lines:
        
        if line in input_lines or line in input_lines.replace(' ', ''):
            matching_percentage += 1
    
    percentage = float(100 * matching_percentage / len(check_lines))
    if percentage > 20:
        return True
    else:
        return False


def get_template_for_vars(template_file):
    """
    This function parses:
    a. all templates located in /templates folder,
    b. the variables of each template in "for loop" (i.e. observable_id's)
    
    It then updates the database "Templates" table with all available templates.
    This function is useful for "viewing" purposes.
    A user can submit a query to the web portal and find all available templates along with the
    corresponding observable_id's
    
    :return:
    """
    for_loop_vars = re.compile(r"{%for (.*?) in .*?%}")
    
    text = template_file.read()
    match = re.search(for_loop_vars, text)
    if match:
        template_for_arguments = re.findall(for_loop_vars, text)[0]
        return template_for_arguments
    return None


def evaluate_variable_part(variable):
    """
    This function evaluates a variable (check parameter).
    :param variable:{0-n}
    :return: ['{0-n}': [0, 1, 2, ..., n]]
    """
    initial_var = copy.deepcopy(variable)
    variable = variable.strip('{}').split('-')
    if len(variable) > 2 or len(variable) < 2:
        utilities_logger.error(
            'Range %s is not correct in the correct format {starting_int - ending_int}' % initial_var)
        raise Exception('Range %s is not correct in the correct format {starting_int - ending_int}' % initial_var)
    
    starting_number = variable[0]
    ending_number = variable[1]
    
    # We need this one to set the argument of "format" function
    # If we are given 01-03, we should return 01, 02, 03
    # and NOT 1, 2, 3
    
    length = len(starting_number)
    try:
        starting_number = int(starting_number)
        # range() is not inclusive!
        ending_number = int(ending_number) + 1
    except:
        utilities_logger.error("Can't convert variables to integers %s, %s" % (starting_number, ending_number))
        raise Exception("Can't convert variables to integers")
    
    the_range = range(starting_number, ending_number)
    # e.g. format(x,'02d')
    format_argument = '0' + str(length) + 'd'
    temp_list = [format(x, format_argument) for x in the_range]
    temp_as_a_dict = dict()
    temp_as_a_dict[initial_var] = temp_list
    
    return temp_as_a_dict


def generate_uri(uri: str(), static_variables=None):
    """
    This function generates a URI derivatives out of a URI which contains a variable part (i.e. {number-number})
    Variable parts are identified with regex.
    A URI may contain more than one variable parts (e.g. http://example.com/{01-02}-{2010-2011}).
    Function should parse them both and generate variable parts **seperately**.
    For the above example, output is:
    - http://example.com/01-2010
    - http://example.com/02-2010
    - http://example.com/01-2011
    - http://example.com/02-2011
    
    Attention: There may be cases in which we have the **very same** variable part more than one times.
    All these parts should be generated **all-together** (and not seperately).
    Consider the following example: http://www.example.com/{2016-2017}{01-02}/IDCJDW2006.{2016-2017}{01-02}
    We can identify to "clusters": {2016-2017} and {01-02} occuring twice each.
    
    For the above example, output is:
    - http://www.example.com/201601/IDCJDW2006.201601
    - http://www.example.com/201701/IDCJDW2006.201701
    - http://www.example.com/201602/IDCJDW2006.201602
    - http://www.example.com/201702/IDCJDW2006.201702
    
    This version supports only for numerical operations.
    
    It also exports a list of uris based on a static input.
    A URI contains a placeholder (i.e. http://example.com/{$var}).
    User defines also a set of vars ("one, two, three, four").
    
    For the above example, output is:
    - http://example.com/one
    - http://example.com/two
    - http://example.com/three
    - http://example.com/four
    
    :param uri: URI with variable member as indicated by user (e.g. http://example.com/{01-09})
    :param static_variables:
    :return: A list with all resulting URIs. In case URI does not contain variable part (i.e. {01-09}),
    it returns one uri in a list
    """
    # variable_regex is more generic. It matches both {01-09} AND {$var} (i.e. static_variable_regex)
    variable_regex = r"({.*?})"
    static_variable_regex = r"({_var_})"
    match = re.search(variable_regex, uri)
    uris_in_a_list = list()
    if match:
        if re.search(static_variable_regex, uri):
            try:
                # We are in the case we have static variables
                static_variables_in_a_list = static_variables.split(',')
                for static_var in static_variables_in_a_list:
                    iteration_uri = copy.deepcopy(uri)
                    iteration_uri = iteration_uri.replace("{_var_}", static_var.strip())
                    uris_in_a_list.append(iteration_uri)
            except AttributeError:
                utilities_logger.error('--extra parameter was not given (Station variables). generate_uri()')
                raise AttributeError('--extra parameter (static vars) was empty')
        else:
            # We want all unique variable parts. See documentation about clusters!
            variable_parts_in_a_list = list(set(re.findall(variable_regex, uri)))
            evaluated_variable_parts = list(map(evaluate_variable_part, variable_parts_in_a_list))
            
            to_be_replaced = list()
            to_replace_with = list()
            for dict_pair in evaluated_variable_parts:
                key, value = dict_pair.popitem()
                to_be_replaced.append(key)
                to_replace_with.append(value)
            
            to_replace_with = list(itertools.product(*to_replace_with))
            
            for item in to_replace_with:
                iteration_uri = copy.deepcopy(uri)
                for count, value in enumerate(item):
                    iteration_uri = iteration_uri.replace(to_be_replaced[count], value)
                uris_in_a_list.append(iteration_uri)
    else:
        uris_in_a_list.append(uri)
    
    return uris_in_a_list


def download_and_check_with_tmpl_html_content_via_http(url: list(), template, template_object,
                                                       storage_type=StorageType.MEMORY):
    """
    This functions downloads and temporarily stores (in memory) html content of a given uri.
    Consequently, it checks if it matches against a given template (check parameter template).
    It the match is successful, html content is permanently stored in a file (in inputs/ folder).
    
    :param url:
    :param template:
    :param storage_type: Whether downloaded files to be stored in a file or in memory
    :return: (input_list, template)
    """
    input_list = list()
    template_name, _ = os.path.splitext(os.path.split(template)[1])
    
    # Create a folder to add downloaded files!
    if storage_type is StorageType.FILE:
        directory = os.path.join(home_directory, 'inputs', template_name)
        try:
            os.makedirs(directory)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
    for count, unique_url in enumerate(url):
        r = requests.get(unique_url)
        if r.status_code == 200:
            data_input = r.text
            if check_template_against_input_file(template_file=template_object, data_input=io.StringIO(data_input)):
                if storage_type is StorageType.FILE:
                    # Keep a backup!
                    file_type = r.headers['content-type']
                    file_encoding = None
                    if 'text/plain' in file_type:
                        file_encoding = ".txt"
                    elif 'text/html' in file_type:
                        file_encoding = ".html"
                    # elif "application/octet-stream" in file_type:
                    else:
                        # TODO: Fix this. It's like this to work for Bom!!!
                        file_encoding = ".txt"
                    filename = os.path.join(home_directory, 'inputs', template_name,
                                            template_name + str(count) + file_encoding)
                    with open(filename, 'w') as fp:
                        fp.write(data_input)
                
                input_list.append(io.StringIO(data_input))
        else:
            utilities_logger.warning(f"There is an error with: {unique_url}")
    
    return input_list, template


def check_if_path_exists(filename=None):
    """
    This utility function checks if a file exists.
    It first checks if the filename contains the path (e.g. ~/user/this/file.txt).
    If not it searches in the three folders located in user's home path (e.g. ~/.edam/inputs/, etc.)
    
    :param filename: The name of the file (e.g. Yucheng.met)
    :return: exists(True/False), filetype(Enum), FileObject (or None)
    """
    if os.path.exists(filename):
        # It means user gave full path
        # It may be a file or a folder
        if os.path.isfile(filename):
            return True, InputType.FILE, filename, open(filename, 'r')
        else:
            return True, InputType.FOLDER, filename, None
    else:
        # It means we have to find it
        for folder, included_folder, filenames in os.walk(home_directory):
            if '.viewer' in included_folder:
                included_folder.remove('.viewer')
            temp_path = os.path.join(folder, filename)
            
            if os.path.exists(temp_path):
                # It means user gave full path
                # It may be a file or a folder
                
                if os.path.isfile(temp_path):
                    return True, InputType.FILE, temp_path, open(temp_path, 'r')
                else:
                    return True, InputType.FOLDER, temp_path, None
        
        # If we are here it means that we couldn't find the file anywhere.
        # We should return False, None, None
        return False, None, None, None


def determine_storage_type(storage_as_string):
    """
    It transform string input ('file' or 'memory') to enum.StorageType
    :param storage_as_string: str: ('file' or 'memory')
    :return: enum.StorageType
    """
    try:
        return StorageType(storage_as_string)
    except:
        utilities_logger.error('Storage option should be \'file\' or \'memory\'. Not %s' % storage_as_string)
        raise SystemExit("Wrong storage option")


def identify_input_type(input_file: str, template, template_object, sql_query=None, extra_variables=None,
                        storage=StorageType.MEMORY):
    """
    This function identifies the type of given input and acts accordingly.
    Possible input type:
    - file
    - folder
    - http_url
    - database_connection string
    
    
    :return:
    """
    available_databases = ['postgres', 'sqlite', 'mysql', 'oracle']
    
    verified_inputs_as_list = []
    file_type = None
    if input_file.__contains__('http'):
        # We have a url!
        
        temp_inputs_as_list = generate_uri(uri=input_file, static_variables=extra_variables)
        
        verified_inputs_as_list, _ = download_and_check_with_tmpl_html_content_via_http(temp_inputs_as_list,
                                                                                        template=template,
                                                                                        template_object=template_object,
                                                                                        storage_type=storage)
        file_type = InputType.HTTP
    elif any(x in input_file for x in available_databases):
        # We have a database connection string
        # TODO: Check if database exists!
        db = records.Database(db_url=input_file)
        rows = db.query(sql_query)
        csv_name = home_directory + '/inputs/' + 'db_dump_' + str(int(datetime.now().timestamp())) + '.csv'
        with open(csv_name, 'w') as fp:
            fp.write(rows.export('csv'))
        if check_template_against_input_file(template_file=template, data_input=csv_name):
            verified_inputs_as_list.append(csv_name)
            file_type = InputType.DATABASE
        else:
            pass
    
    else:
        # We must have a file or a folder!
        exists, typ, input_file_path, _ = check_if_path_exists(input_file)
        if exists:
            if typ is InputType.FOLDER:
                # folder! Check if exists and iterate through all available files
                directory = input_file_path
                potential_files = list()
                for folder, included_folder, filenames in os.walk(directory):
                    for filename in filenames:
                        if not fnmatch.fnmatch(filename, '*.tmpl') and not fnmatch.fnmatch(filename, '.*'):
                            potential_files.append(open(os.path.join(folder, filename), 'r'))
                
                for potential_file in potential_files:
                    if check_template_against_input_file(template_file=template_object, data_input=potential_file):
                        verified_inputs_as_list.append(potential_file)
                        file_type = InputType.FILE
            
            else:
                # File
                # First we have to check if we can generate uris (e.g. files is "uk{01-90}.txt")
                # temp_list is either ['uk.txt'] (if input_file was "uk.txt") or
                # ['uk1.txt', 'uk2.txt', ..] (if input_file was "uk{01-90}.txt")
                temp_list = generate_uri(input_file_path, static_variables=extra_variables)
                for generated_file_uri in temp_list:
                    exists, file_type, path, file_object = check_if_path_exists(filename=generated_file_uri)
                    if exists and file_type is InputType.FILE:
                        # Let's check if we verify it with the template!
                        if check_template_against_input_file(template_file=template_object, data_input=file_object):
                            verified_inputs_as_list.append(file_object)
                            file_type = InputType.FILE
        else:
            raise SystemExit("%s does not exist" % input_file)
    if not verified_inputs_as_list:
        # It means we didn't find a single input to match with a template
        utilities_logger.warning("identify_input_type(): No verified inputs found")
        utilities_logger.info("No verified inputs found")
        return False, None, None
    else:
        return True, verified_inputs_as_list, file_type

if __name__ == "__main__":
    pass


import io
import os

import pandas as pd
import yaml
from geopy.geocoders import GoogleV3

from edam.reader.manage import DatabaseHandler
from edam.reader.models import Station, Sensors, AbstractObservables, UnitsOfMeasurement, \
    HelperTemplateIDs
from edam.reader.utilities import parse_for_iterations, extract_data_from_preamble, check_if_path_exists


def safe_float(value):
    try:
        return float(value)
    except (ValueError, TypeError):
        return None


class SourceConfiguration:
    """
    This class handles configuration files, drafted by users.
    It reads it and extracts all relevant information.
    1. Creates the Station and Sensors objects. It checks if Station exists and appends related Sensors in it.
    2. Sets file locators for all corresponding input and output templates
    3. Sets locators for all corresponding inputs. Depending on the type (html, file, sql), calls the appropriate
    connector of the Connector class.
    
    A SourceConfiguration object serves as an input to TemplateReader.
    """
    
    def __init__(self, input_yaml, input_file_data=io.StringIO(), input_preamble=io.StringIO(),
                 template_preamble=io.StringIO()):
        """
        """
        # TODO: Implemenet this one..
        self.sensor_id = None
        
        self.database = DatabaseHandler()
        
        self.input_yaml = input_yaml
        self.input_file = input_file_data
        self.input_preamble = input_preamble
        self.template_preamble = template_preamble
        
        self.input_yaml.seek(0)
        self.input_file.seek(0)
        self.input_preamble.seek(0)
        self.template_preamble.seek(0)
        
        self.helper_template = pd.DataFrame(columns=['observable_id',
                                                     'abstract_observable_id',
                                                     'unit_id',
                                                     'station_id',
                                                     'sensor_id'])
        # self.available_fields = ['Station', 'Observables', 'Units of Measurement', 'Sensors', 'Data inputs']
        self.station_id = []
        self.content = None
        self.all_observable_ids = list()
        
        self.handler()
    
    def check_yaml(self):
        try:
            return yaml.load(self.input_yaml, Loader=yaml.FullLoader)
        except yaml.YAMLError as exc:
            return exc
    
    def handler(self):
        # Firstly we open the yaml
        self.content = self.check_yaml()
        
        # Check Station type
        station_type, source, template = self.__check_type_of_field(field_name='Station')
        
        if station_type == 'iterable':
            self.station_id = parse_for_iterations(input_iteration_file=source,
                                                   template_iteration_file=template, iterable_type='Station')
        else:
            self.set_station(parse_from_yaml=station_type)
        # Do the same for all other fields
        # Though this could be done only once.
        observable_type, source, template = self.__check_type_of_field(field_name='Observables')
        if observable_type == 'iterable':
            self.set_observables(iter_or_not=parse_for_iterations(input_iteration_file=source,
                                                                  template_iteration_file=template,
                                                                  iterable_type='Observables'))
        else:
            self.set_observables()
        
        # We should take care cases in which uom and sensor fields are empty.
        # If these fields are empty, it means that they should default to "unknown values"
        self.set_units_of_measurement()
        self.set_sensors()
        
        # Update dataframe. For some reason int is transformed to float. So here I revert this (for the affected columns
        self.helper_template['station_id'] = self.helper_template['station_id'].apply(int)
        self.helper_template['sensor_id'] = self.helper_template['sensor_id'].apply(int)
        self.helper_template['abstract_observable_id'] = self.helper_template['abstract_observable_id'].apply(int)
        self.helper_template['unit_id'] = self.helper_template['unit_id'].apply(int)
        
        # Now copy this dataframe so many times as the len(self.station_id)
        # self.helper_template = self.helper_template.append(temp, ignore_index=True)
        # We want to pass the first station_id since we already have this id incorporated.
        temp = self.helper_template.copy(deep=True)
        for station_id in self.station_id[1:]:
            temp['station_id'] = station_id
            self.helper_template = self.helper_template.append(temp, ignore_index=True)
        
        del temp
        # We have to check if we have duplicates!
        df_cleaned = self.database.clean_df_db_dups(df=self.helper_template, tablename='HelperTemplateIDs',
                                                    dup_cols=list(self.helper_template))
        
        self.database.__add_dataframe__(dataframe=df_cleaned, table='HelperTemplateIDs', index=False)
    
    def set_station(self, parse_from_yaml):
        """
        :return:
        """
        # With the following command I serialize a Station object from the .yaml file
        if parse_from_yaml is not None:
            station = Station.fromdictionary(self.content['Station'])  # type: Station
        else:
            station = Station()
        
        # It means metadata have to be parsed from the preambles
        if self.input_preamble.seek(0, os.SEEK_END) > 0 and self.template_preamble.seek(0, os.SEEK_END) > 0:
            self.input_preamble.seek(0)
            self.template_preamble.seek(0)
            station = extract_data_from_preamble(station, preamble_template=self.template_preamble
                                                 , preamble_input=self.input_preamble)
        
        # With the following command, I determine the existence of station object
        # with the same attributes (non-duplicate entries)
        
        exists, station_from_db = self.database.__check_station_is_in_db__(station)
        if exists:
            station_id = station_from_db.id
        else:
            station.latitude = safe_float(station.latitude)
            station.longitude = safe_float(station.longitude)
            if station.latitude is None and station.longitude is None and station.name is not None:
                geolocator = GoogleV3()
                try:
                    location = geolocator.geocode(station.name + station.region)
                except:
                    try:
                        location = geolocator.geocode(station.name)
                    except:
                        location = None
                if location is not None:
                    station.latitude = location.latitude
                    station.longitude = location.longitude
            
            _, station_id = self.database.__add_item__(station)
        
        self.station_id.append(station_id)
    
    def set_observables(self, iter_or_not=None):
        
        if iter_or_not is None:
            observables = self.content['Observables']
        else:
            observables = iter_or_not
            
            # This is where we should parse metadata of observables
            # parse_observables_with_reasoner(observables=observables)
        for obs in observables:
            observable_as_dict = obs  # type: dict
            # Deprecated
            # observable_as_dict['station_id'] = self.station_id
            observable = AbstractObservables.fromdictionary(observable_as_dict)
            exists, respective_abstract_observable_id = self.database.__check_observable_is_in_db__(observable)
            if exists:
                respective_abstract_observable_id = respective_abstract_observable_id[0]
            else:
                _, respective_abstract_observable_id = self.database.__add_item__(observable)
            
            # Create the 1/(len(station_id)) dataframe. The others would be exactly the same
            # apart the station_id section. This is derived of cedar requirements. I.e. Observables, uoms, sensors, etc
            # located in a config file regard ALL THE STATIONS in the config.
            
            temp = pd.Series({'observable_id': observable_as_dict['observable_id'],
                              'abstract_observable_id': respective_abstract_observable_id,
                              'unit_id': None,
                              'station_id': self.station_id[0],
                              'sensor_id': self.sensor_id
                              })
            
            self.all_observable_ids.append(observable_as_dict['observable_id'])
            
            self.helper_template = self.helper_template.append(temp, ignore_index=True)
    
    def set_helper_observable_ids(self, helper_template_as_dictionary):
        helperTemplateID = HelperTemplateIDs.fromdictionary(helper_template_as_dictionary)
        exists, _ = self.database.__chech_helperTemplateID_is_in_db__(helperTemplateID)
        if exists:
            pass
        else:
            _, _ = self.database.__add_item__(helperTemplateID)
    
    def set_units_of_measurement(self):
        
        if self.content['Units of Measurement'] is None:
            default_empty_uom = dict()
            default_empty_uom['name'] = "unknown"
            relevant_observables = self.all_observable_ids
            
            unit = UnitsOfMeasurement.fromdictionary(default_empty_uom)
            
            exists, unit_id = self.database.__check_unit_is_in_db__(unit)
            if exists:
                unit_id = unit_id[0]
            else:
                _, unit_id = self.database.__add_item__(unit)
            pass
            
            for observable_observable_id in relevant_observables:
                self.helper_template.loc[
                    self.helper_template['observable_id'] == observable_observable_id, 'unit_id'] = unit_id
        
        else:
            for uom in self.content['Units of Measurement']:
                uom_as_dict = uom  # type: dict
                if uom_as_dict['relevant_observables'] == '':
                    relevant_observables = self.all_observable_ids
                else:
                    relevant_observables = uom_as_dict['relevant_observables'].split(',')  # type: list
                    # remove spaces
                    relevant_observables = map(str.strip, relevant_observables)
                    # No need to keep it any more
                del uom_as_dict['relevant_observables']
                unit = UnitsOfMeasurement.fromdictionary(uom_as_dict)
                
                exists, unit_id = self.database.__check_unit_is_in_db__(unit)
                if exists:
                    unit_id = unit_id[0]
                else:
                    _, unit_id = self.database.__add_item__(unit)
                
                for observable_observable_id in relevant_observables:
                    self.helper_template.loc[
                        self.helper_template['observable_id'] == observable_observable_id, 'unit_id'] = unit_id
    
    def set_sensors(self):
        """
        :return:
        """
        if self.content['Sensors'] is None:
            default_empty_sensor = dict()
            default_empty_sensor['generic'] = True
            
            relevant_observables = self.all_observable_ids
            for observable_observable_id in relevant_observables:
                sensor = Sensors.fromdictionary(default_empty_sensor)
                # abstract_observable_id = None or id
                abstract_observable_id = \
                    self.helper_template.loc[(self.helper_template['station_id'] == self.station_id[0]) &
                                             (self.helper_template['observable_id'] == observable_observable_id)][
                        'abstract_observable_id'].values[0]
                abstract_observable_id = int(abstract_observable_id)
                
                unit_id = \
                    self.helper_template.loc[(self.helper_template['station_id'] == self.station_id[0]) &
                                             (self.helper_template['observable_id'] == observable_observable_id)][
                        'unit_id'].values[0]
                unit_id = int(unit_id)
                
                default_empty_sensor['unit_id'] = unit_id
                default_empty_sensor['abstract_observable_id'] = abstract_observable_id
                sensor.update(default_empty_sensor)
                
                exists, sensor_id = self.database.__check_sensor_is_in_db__(sensor)
                
                if exists:
                    sensor_id = sensor_id[0]
                else:
                    _, sensor_id = self.database.__add_item__(sensor)
                
                # We now need to update helper template ids table
                self.helper_template.loc[
                    (self.helper_template['station_id'] == self.station_id[0]) &
                    (self.helper_template['observable_id'] == observable_observable_id), 'sensor_id'] = sensor_id
        
        else:
            
            for sensor in self.content['Sensors']:
                sensor_as_dict = sensor  # type: dict
                if sensor_as_dict['relevant_observables'] == '':
                    relevant_observables = self.all_observable_ids
                else:
                    relevant_observables = sensor_as_dict['relevant_observables'].split(',')  # type: list
                    relevant_observables = map(str.strip, relevant_observables)
                    # No need to keep it any more
                # No need to keep it any more
                del sensor_as_dict['relevant_observables']
                # We have to retrieve abstract_observable_id from observable_id
                # We are going to do this, through the HelperTemplateIDs
                # After retrieving this id, we will update 'sensor' object, check if it's in db already
                # And finally store it.
                
                # A generic sensor can have more than one relevant_observables
                # In this case, we are going to create as many generic sensor objects as the relevant_observables..
                
                # We are going to determine the observable_id through the observable_id
                
                for observable_observable_id in relevant_observables:
                    sensor = Sensors.fromdictionary(sensor_as_dict)
                    # abstract_observable_id = None or id
                    abstract_observable_id = \
                        self.helper_template.loc[(self.helper_template['station_id'] == self.station_id[0]) &
                                                 (self.helper_template['observable_id'] == observable_observable_id)][
                            'abstract_observable_id'].values[0]
                    abstract_observable_id = int(abstract_observable_id)
                    
                    unit_id = \
                        self.helper_template.loc[(self.helper_template['station_id'] == self.station_id[0]) &
                                                 (self.helper_template['observable_id'] == observable_observable_id)][
                            'unit_id'].values[0]
                    unit_id = int(unit_id)
                    
                    sensor_as_dict['unit_id'] = unit_id
                    sensor_as_dict['abstract_observable_id'] = abstract_observable_id
                    sensor.update(sensor_as_dict)
                    sensor.generic = True
                    exists, sensor_id = self.database.__check_sensor_is_in_db__(sensor)
                    # Next line is resolving a bug introduced by sqlite
                    # For some unknown reason a str type was interpreted as dict when it came for storing
                    sensor.tags = str(sensor.tags)
                    if exists:
                        sensor_id = sensor_id[0]
                    else:
                        _, sensor_id = self.database.__add_item__(sensor)
                    
                    # We now need to update helper template ids table
                    self.helper_template.loc[
                        (self.helper_template['station_id'] == self.station_id[0]) &
                        (self.helper_template['observable_id'] == observable_observable_id), 'sensor_id'] = sensor_id
    
    def __check_type_of_field(self, field_name):
        """
        This function checks each given field values and determines the "type" of values.
        These could be in the form of "source:..., template:...", which means we need to
        iterate through source file and extract the fields. The other type is where fields
        are manually iterated in the config file, in the form of 1..., 2..., 3...
        :return: source_type (iterable, non_iterable), source (source_path, None), template (template_path, None)
        """
        # If True it means we have to extract data from iterable.
        try:
            if ('source' and 'template') in self.content[field_name]:
                source_type = 'iterable'
                source = self.content[field_name]['source']
                template = self.content[field_name]['template']
                sexists, _, sourcef, source_io_object = check_if_path_exists(source)
                texists, _, templatef, template_io_object = check_if_path_exists(template)
                
                if sexists and texists:
                    pass
                else:
                    # TODO logging!
                    raise SystemExit("%s and %s does not exist" % (source, template))
            
            else:
                source_type = 'non_iterable'
                source_io_object = io.StringIO()
                template_io_object = io.StringIO()
            
            return source_type, source_io_object, template_io_object
        except KeyError:
            # It means that yaml does not contain this field (e.g. Station, or Observable).
            # That is because metadata are PROBABLY stored in input files (preamble)
            return None, None, None


if __name__ == "__main__":
    test = SourceConfiguration('metadata/knmi.yaml', 'inputs/Yucheng.csv')



import calendar
import hashlib
import os
import shutil
from datetime import datetime
from functools import wraps, update_wrapper

import jinja2
import pandas as pd
from flask import Flask, url_for, redirect
from flask import make_response
from flask_caching import Cache
from flask_sqlalchemy import SQLAlchemy

import edam.viewer.config as config
from edam.settings import home_directory
from edam.viewer.app.manage import DatabaseHandler as Data
from edam.viewer.app.manage import Measurement
from edam.viewer.app.utilities import check_template_source_compatibility


def copytree(src, dst, symlinks=False, ignore=None):
    for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.isdir(s):
            shutil.copytree(s, d, symlinks, ignore)
        else:
            if "DS_" not in s:
                shutil.copy2(s, d)


app = Flask(__name__, static_folder=os.path.join(home_directory, '.viewer/', 'static'),
            static_url_path=os.path.join(home_directory, '.viewer/'))
app.config.from_object(config)

db = SQLAlchemy(app)

# Jinja trim whitespace
app.jinja_env.trim_blocks = True
app.jinja_env.lstrip_blocks = True

# Every time the server starts copy all user templates for viewing purposes
# TODO: It would be more efficient to create symlinks than hard-copying files
# copytree(os.path.join(home_directory, 'templates'),
#          os.path.join(home_directory, '.viewer', 'templates', 'edam'))

# Change the folder of templates and static files
app.jinja_loader = jinja2.FileSystemLoader([home_directory + '/.viewer/templates'])

# Don't sort with jsonify
app.config['JSON_SORT_KEYS'] = False

# Set cache
cache = Cache(app, config={'CACHE_TYPE': 'simple'})
cache.init_app(app)

# secret key for sessions
app.secret_key = "this should be harder to guess"

data = Data()


@cache.cached(timeout=600, key_prefix='metastations')
def calculate_metastations():
    metastations = data.retrieve_stations()
    return metastations


@cache.cached(timeout=600, key_prefix='metatemplates')
def calculate_templates():
    metatemplates = data.retrieve_templates()
    return metatemplates


def calculate_data_and_render_from_template(observable_id, station_id):
    station_object = data.retrieve_object_from_id(table='Station', object_id=int(station_id))  # type: models.Station
    compatible, list_template_for_arguments, template_dictionary = check_template_source_compatibility(
        observable_id=observable_id,
        station_object=station_object)
    if compatible:
        station, chunk = data.retrieve_stations_data(station_object, list_template_for_arguments)
        return True, template_dictionary, station, chunk
    else:
        return False, redirect(url_for('index')), None, None


def nocache(view):
    @wraps(view)
    def no_cache(*args, **kwargs):
        response = make_response(view(*args, **kwargs))
        response.headers['Last-Modified'] = datetime.now()
        response.headers['Cache-Control'] = 'no-store, no-cache, must-revalidate, post-check=0, pre-check=0, max-age=0'
        response.headers['Pragma'] = 'no-cache'
        response.headers['Expires'] = '-1'
        return response

    return update_wrapper(no_cache, view)


def flot(timestamp):
    return calendar.timegm(timestamp.timetuple()) * 1000


app.jinja_env.globals.update(flot=flot)


def hash(anything):
    return '_' + hashlib.md5(anything).hexdigest()


app.jinja_env.globals.update(hash=hash)


def d2s(datetime):
    return datetime.strftime("%D %H %M")


app.jinja_env.globals.update(d2s=d2s)


def same_timestamp(*args):
    if args:
        return args[0]
    else:
        return None


app.jinja_env.globals.update(same_timestamp=same_timestamp)


def resample(df: pd.DataFrame, rule, how=None, axis=0, fill_method=None, closed=None, label=None, convention='start',
             kind=None, loffset=None, limit=None, base=0, on=None, level=None):
    # observables_list = list(df)
    # pd.set_option('precision', 3)
    observables_list = ['timestamp', 'tmax', 'tmin', 'af', 'rain', 'sun']
    # observables_list = ['timestamp', 'radn', 'maxt', 'mint', 'rain', 'wind', 'RH']
    observables_list.remove('timestamp')
    available_operations = ['bfill', 'max', 'median', 'sum', 'min', 'interpolate', 'ffill']

    try:
        for observable in observables_list:
            df[observable] = df[observable].apply(lambda x: float(x))

    except Exception as e:
        print(e.args)
        print("I can't transform string value to float. Wind maybe? Check edam.viewer.__init__.py - downsample func")
        exit()
    resampled = df.resample("A", axis, fill_method, closed, label, convention, kind, loffset, limit, base, on,
                            level)

    resampled = resampled.mean()
    resampled = resampled.round(3)
    resampled = resampled.fillna('---')

    if how is None:
        if how in available_operations:
            # resampled = getattr(resampled, "interpolate")(method)
            resampled = getattr(resampled, how)()
    resampled["timestamp"] = resampled.index

    for observable in observables_list:
        resampled[observable] = resampled[observable].apply(lambda x: Measurement(x))
    # TODO: This is soooooo dangerous. Please re-implement......
    # observables_list.append('timestamp')
    observables_list = ['timestamp', 'tmax', 'tmin', 'af', 'rain', 'sun']
    # observables_list = ['timestamp', 'radn', 'maxt', 'mint', 'rain', 'wind', 'RH']
    zip_argument = map(lambda x: "resampled." + x, observables_list)

    zip_argument = ",".join(zip_argument)

    zip_argument = eval("zip(%s)" % zip_argument)

    return zip_argument


app.jinja_env.globals.update(resample=resample)

# def jsoni(uni):
#     return json.loads(uni, object_hook=json_util.object_hook)
# app.jinja_env.globals.update(jsoni = jsoni)



from edam.reader.models import Station

from edam.reader.utilities import find_and_describe_template


def check_template_source_compatibility(observable_id, station_object: Station):
    template_dictionary = find_and_describe_template(observable_id)
    print(observable_id)
    if template_dictionary:
        template_for_vars = template_dictionary['variables'].replace(' ', '').split(',')
    else:
        return False, None, False
    station_observable_ids = [helper.observable_id for helper in station_object.helper]
    station_observable_ids.append('timestamp')
    if set(template_for_vars) <= set(station_observable_ids):
        return True, template_for_vars, template_dictionary
    else:
        return False, None, None



import errno
import io
import os
import shutil
import sys

from setuptools import setup, find_packages
from setuptools.command.install import install
from setuptools.command.test import test as TestCommand

__version__ = '1.0.40'
here = os.path.abspath(os.path.dirname(__file__))


def copytree(src, dst, symlinks=False, ignore=None):
    for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.isdir(s):

            try:
                shutil.copytree(s, d, symlinks, ignore)
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise
        else:
            shutil.copy2(s, d)


# https://stackoverflow.com/questions/15853058/run-custom-task-when-call-pip-install
class CustomInstall(install):
    def run(self):
        install.run(self)
        # custom stuff here
        # create folders to put staff in.
        directories_to_be_created = [os.path.expanduser("~/.edam/"), os.path.expanduser("~/.edam/templates/"),
                                     os.path.expanduser("~/.edam/metadata/"),
                                     os.path.expanduser("~/.edam/inputs/"),
                                     os.path.expanduser("~/.edam/.viewer/")
                                     ]
        for directory in directories_to_be_created:
            try:
                os.makedirs(directory, exist_ok=True)
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise

        resources_directory = os.path.join(here, 'edam', 'resources')
        home_user_directory = os.path.expanduser("~/.edam/")
        directories_to_be_copied_from_resources = ['inputs', 'templates', 'metadata']

        for directory in directories_to_be_copied_from_resources:
            copytree(os.path.join(resources_directory, directory), os.path.join(home_user_directory, directory))
        shutil.copyfile(os.path.join(resources_directory, 'settings.yaml'),
                        os.path.join(home_user_directory, 'settings.yaml'))

        shutil.copyfile(os.path.join(resources_directory, 'edam.owl'),
                        os.path.join(home_user_directory, 'edam.owl'))

        shutil.copyfile(os.path.join(resources_directory, 'edam.owl'),
                        os.path.join(home_user_directory, 'backup.owl'))
        # Copy flask_related contents into home_directory

        copytree(os.path.join(resources_directory, 'flask_related'),
                 os.path.join(home_user_directory, '.viewer/'))
        # Copy edam templates into flask edam/templates
        copytree(os.path.join(resources_directory, 'templates'),
                 os.path.join(home_user_directory, '.viewer', 'templates', 'edam'))


def read(*filenames, **kwargs):
    encoding = kwargs.get('encoding', 'utf-8')
    sep = kwargs.get('sep', '\n')
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


long_description = read('README.md')


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        errcode = pytest.main(self.test_args)
        sys.exit(errcode)


setup(
    name='edam',
    version=__version__,
    url='https://github.com/BigDataWUR/EDAM',
    description='An input template framework',
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='GNU General Public License v3.0',
    author='Argyrios Samourkasidis',
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    include_package_data=True,
    install_requires=['sqlalchemy', 'requests', 'numpy', 'pandas', 'numexpr', 'geopy', 'Flask-SQLAlchemy',
                      'Flask', 'Flask-Caching', 'jinja2', 'PyYAML==5.4.1', 'records', 'psycopg2', 'click',
                      'Flask-GoogleMaps==0.2.4', 'owlready2', 'pint', 'Werkzeug'
                      ],
    cmdclass={'test': PyTest, 'install': CustomInstall},
    python_requires='>=3.6',
    packages=find_packages(exclude=["tests.*", "tests"]),
    author_email='argysamo@gmail.com',
    package_data={
        'edam': ['resources/metadata/*', 'resources/inputs/*', 'resources/templates/*',
                 'resources/settings.yaml', 'resources/flask_related/static/*/*',
                 'resources/flask_related/templates/*/*', ],
    },
    entry_points={
        'console_scripts':
            ['edam=bin.read:cli', 'viewer=bin.viewer:run'],

    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],

)



import logging
from io import StringIO

import pandas as pd
from sqlalchemy import and_
from sqlalchemy import create_engine
from sqlalchemy.orm import Query
from sqlalchemy.orm import scoped_session, sessionmaker

from edam.reader.models import Sensors, Station, Observations, \
    Base, AbstractObservables, UnitsOfMeasurement, HelperTemplateIDs
from edam.settings import database_url, database_type

module_logger = logging.getLogger('edam.reader.manage')


def db_connect():
    """
    Performs database connection using database settings from settings.py.
    Returns sqlalchemy engine instance
    """
    module_logger.debug(f'Received a call to `db_connect()`. Database url is: {database_url}')
    return create_engine(database_url)


db_session = scoped_session(sessionmaker(autocommit=False,
                                         autoflush=False,
                                         bind=db_connect()))


def create_tables(engine):
    """"""
    Base.metadata.create_all(engine)


def drop_tables(engine):
    module_logger.debug('Received a call to `drop_tables()`.')
    module_logger.info('Database tables dropped using drop_tables()')
    Base.metadata.drop_all(engine)


class DatabaseInstantiation(object):
    def __init__(self, drop=True):
        """
        Initializes database connection and sessionmaker.
        """
        db_insta_logger = logging.getLogger('edam.reader.manage.DatabaseInstantiation')
        self.engine = db_connect()
        db_insta_logger.debug('Finding database engine connection string')
        if drop:
            db_insta_logger.debug('Calling drop_tables()')
            drop_tables(self.engine)
        db_insta_logger.debug('Creating database')
        create_tables(self.engine)


class DatabaseHandler(object):
    """
    This class handles database operations (add item, add dataframes, etc.)
    """

    def __init__(self):
        """
        Initializes database connection and sessionmaker.
        """
        self.db_handler_logger = logging.getLogger('edam.reader.manage.DatabaseHandler')

        self.engine = db_connect()

        self.Session = db_session

    def clean_df_db_dups(self, df, tablename, dup_cols=None, filter_continuous_col=None, filter_categorical_col=None):
        """
        code: https://github.com/ryanbaumann/Pandas-to_sql-upsert/blob/master/to_sql_newrows.py
        Remove rows from a dataframe that already exist in a database
        Required:
            df : dataframe to remove duplicate rows from
            engine: SQLAlchemy engine object
            tablename: tablename to check duplicates in
            dup_cols: list or tuple of column names to check for duplicate row values
        Optional:
            filter_continuous_col: the name of the continuous data column for BETWEEEN min/max filter
                                   can be either a datetime, int, or float data type
                                   useful for restricting the database table size to check
            filter_categorical_col : the name of the categorical data column for Where = value check
                                     Creates an "IN ()" check on the unique values in this column
        Returns
            Unique list of values from dataframe compared to database table
        """
        if dup_cols is None:
            dup_cols = []
        engine = self.engine
        args = 'SELECT {} FROM {}'.format(', '.join(['"{0}"'.format(col) for col in dup_cols]), tablename)
        args_contin_filter, args_cat_filter = None, None
        if filter_continuous_col is not None:
            if df[filter_continuous_col].dtype == 'datetime64[ns]':
                args_contin_filter = f""" "{filter_continuous_col}" BETWEEN Convert(datetime, '{df[filter_continuous_col].min()}')
                                              AND Convert(datetime, '{df[filter_continuous_col].max()}')"""

        if filter_categorical_col is not None:
            args_cat_filter = ' "{}" in({})'.format(filter_categorical_col,
                                                    ', '.join(["'{0}'".format(value) for value in
                                                               df[filter_categorical_col].unique()]))

        if args_contin_filter and args_cat_filter:
            args += ' Where ' + args_contin_filter + ' AND' + args_cat_filter
        elif args_contin_filter:
            args += ' Where ' + args_contin_filter
        elif args_cat_filter:
            args += ' Where ' + args_cat_filter

        df.drop_duplicates(dup_cols, keep='last', inplace=True)
        df2 = pd.read_sql(args, engine)
        for column in list(df):
            if df[column].dtype == "float64":
                df[column] = df[column].astype(str)
                df2[column] = df2[column].astype(str)

        df = pd.merge(df, df2, how='left', on=dup_cols, indicator=True)

        df = df[df['_merge'] == 'left_only']
        df.drop(['_merge'], axis=1, inplace=True)
        if "tation" in tablename:
            if database_type == "postgresql":
                df['id'] = pd.read_sql_query('select coalesce(max(id),0)+1 from public.\"Station\"', engine).iloc[
                               0, 0] + range(len(df))
            elif database_type == "sqlite":
                df['id'] = pd.read_sql_query('select coalesce(max(id),0)+1 from Station', engine).iloc[
                               0, 0] + range(len(df))
            df.set_index(['id'], inplace=True)

        return df

    @staticmethod
    def __add_dataframe__(dataframe: pd.DataFrame, table='Observations', index=True, index_label='timestamp'):
        engine = create_engine(database_url)

        def create_file_object(df, file_path=None, string_io=True, index=index):
            """Creates a csv file or writes to memory"""
            if string_io:
                s_buf = StringIO()
                df.to_csv(s_buf, date_format="%Y-%m-%d %H:%M:%S", index=index)
                s_buf.seek(0)
                file_object = s_buf
            else:
                df.to_csv(file_path, index=False)
                df = open(file_path)
                file_object = df
            return file_object

        def load_to_database(table1, unique_columns, file_object):

            connection = engine.raw_connection()
            try:
                cursor = connection.cursor()

                columns = ', '.join(['{}'.format(col) for col in unique_columns])
                sql = f'COPY "{table1}" ({columns}) FROM STDIN WITH CSV HEADER'
                cursor.copy_expert(sql=sql, file=file_object)

                connection.commit()

                connection.close()
            finally:
                pass

        if database_type == "postgresql":
            df_index_name = dataframe.index.name
            if df_index_name == "id" or df_index_name is None:
                uniq_cols = list(dataframe)
                index = False
            else:
                uniq_cols = list()
                uniq_cols.append(df_index_name)
                uniq_cols += list(dataframe)
                index = True
            load_to_database(table, uniq_cols, create_file_object(dataframe, index=index))
        elif database_type == "sqlite":
            dataframe.to_sql(name=table, con=engine, if_exists='append',
                             index=index, index_label=index_label)

    def __add_item__(self, item):
        """

        :param item:
        :return: True, item.id
        """
        session = self.Session()
        session.expire_on_commit = False
        try:
            session.add(item)
            session.commit()
        except:
            self.db_handler_logger.error(f'Exception when adding {item}. Check __add_item__()')
            session.rollback()
            raise
        finally:
            session.flush()
        session.close()
        return True, item.id

    def __update_item__(self, item, metadata_dict):
        session = self.Session()
        returned_item = session.query(item.__class__).filter(
            item.__class__.id == item.id).first()  # type: HelperTemplateIDs
        returned_item.update_meta(metadata_dict)
        session.commit()
        session.close()

    def __get_observations_by_id_as_df__(self, observable_id):
        """

        :return: A list with all Observations.
        """
        session = self.Session()
        df = pd.DataFrame()
        q = session.query(Observations).filter(Observations.observable_id == observable_id)  # type: Query
        q = q.order_by(Observations.timestamp.asc())
        df = pd.read_sql(sql=q.statement, con=self.engine, index_col='timestamp')
        df.drop('id', axis=1, inplace=True)
        session.rollback()
        session.close()
        return df

    def __check_station_is_in_db__(self, station: Station):
        session = self.Session()

        exists = session.query(Station.id).filter(and_(Station.name == station.name
                                                       , Station.location == station.location
                                                       , Station.mobile == station.mobile)
                                                  )
        session.rollback()
        session.close()
        return exists.scalar() is not None, exists.first()

    def __check_unit_is_in_db__(self, unit: UnitsOfMeasurement):
        session = self.Session()
        exists = session.query(UnitsOfMeasurement.id).filter(and_(UnitsOfMeasurement.name == unit.name
                                                                  , UnitsOfMeasurement.symbol == unit.symbol
                                                                  )
                                                             )
        session.rollback()
        session.close()
        return exists.scalar() is not None, exists.first()

    def __check_observable_is_in_db__(self, observable: AbstractObservables):
        session = self.Session()
        exists = session.query(AbstractObservables.id).filter(and_(AbstractObservables.name == observable.name
                                                                   ,
                                                                   AbstractObservables.ontology == observable.ontology)
                                                              )
        session.rollback()
        session.close()
        return exists.scalar() is not None, exists.first()

    def __chech_helperTemplateID_is_in_db__(self, helperTemplateID: HelperTemplateIDs):
        session = self.Session()
        exists = session.query(HelperTemplateIDs.id). \
            filter(and_(HelperTemplateIDs.observable_id == helperTemplateID.observable_id,
                        HelperTemplateIDs.unit_id == helperTemplateID.unit_id,
                        HelperTemplateIDs.station_id == helperTemplateID.station_id,
                        HelperTemplateIDs.abstract_observable_id == helperTemplateID.abstract_observable_id,
                        HelperTemplateIDs.sensor_id == helperTemplateID.sensor_id))
        session.rollback()
        session.close()
        return exists.scalar() is not None, exists.first()

    def __check_sensor_is_in_db__(self, sensor: Sensors):
        session = self.Session()
        # TODO: Fix the filter parameters. Needs to be more generic
        # TODO: Implement tags equality check (json error)
        exists = session.query(Sensors.id).filter(and_(Sensors.generic == sensor.generic
                                                       , Sensors.abstract_observable_id == sensor.abstract_observable_id
                                                       , Sensors.manufacturer == sensor.manufacturer
                                                       , Sensors.name == sensor.name
                                                       ))
        session.rollback()
        session.close()
        return exists.scalar() is not None, exists.first()

    def __get_helper_table_row_input_file_observable_id__(self, observable_id,
                                                          station_id) -> HelperTemplateIDs:
        """
        
        :param observable_id: This is usually a short of the Observables name, which is used in templates
         E.g. temp is used for Temperature.
        :return:
        """
        session = self.Session()
        q = session.query(HelperTemplateIDs).filter(and_(HelperTemplateIDs.station_id == station_id,
                                                         HelperTemplateIDs.observable_id == observable_id
                                                         )).first()  # type: HelperTemplateIDs
        session.expunge(q)
        # session.rollback()
        session.close()
        return q

    def __get_observable_by__id__(self, id) -> AbstractObservables:
        """

        :param id: This is the id of the stored Observable
         E.g. 1
        :return:
        """
        session = self.Session()

        q = session.query(AbstractObservables).filter(AbstractObservables.id == id)  # type: Query
        session.rollback()
        session.close()
        return q.first()

    def __update_helper_observable_ids_with_unit_id__(self, station_id, observable_observable_id, unit_id):
        # TODO: This should be more generic
        session = self.Session()

        result = session.query(HelperTemplateIDs).filter(and_(HelperTemplateIDs.station_id == station_id,
                                                              HelperTemplateIDs.observable_id == observable_observable_id
                                                              )).first()  # type: HelperTemplateIDs
        # Above query MUST return one result
        # We update this one
        result.unit_id = unit_id

        session.commit()
        session.close()

    def __update_helper_observable_ids_with_sensor_id__(self, station_id, observable_observable_id,
                                                        sensor_id):
        # TODO: This should be more generic
        session = self.Session()

        result = session.query(HelperTemplateIDs).filter(and_(HelperTemplateIDs.station_id == station_id,
                                                              HelperTemplateIDs.observable_id == observable_observable_id
                                                              )).first()  # type: HelperTemplateIDs
        # Above query MUST return one result
        # We update this one
        result.sensor_id = sensor_id

        session.commit()
        session.close()

    def __get_abstract_observable_id_from_observable_id__(self, station_id, observable_observable_id):
        session = self.Session()
        result = session.query(HelperTemplateIDs).filter(and_(HelperTemplateIDs.station_id == station_id,
                                                              HelperTemplateIDs.observable_id == observable_observable_id
                                                              )).first()  # type: HelperTemplateIDs
        session.expunge_all()
        session.close()
        return result.abstract_observable_id or None

    def __get_station_id_by_tags_station_id__(self, tags_station_id):
        session = self.Session()
        string_to_find = "\"station_id\":\"%d\"" % tags_station_id
        exists = session.query(Station.id).filter(Station.tags.contains(string_to_find))

        if exists.first() is not None:
            database_station_id = exists.first().id
        else:
            database_station_id = None
        session.close()
        return database_station_id

    def get_all_observables(self):
        session = self.Session()
        session.close()
        return session.query(AbstractObservables).all()

    def get_all_stations(self):
        session = self.Session()
        session.close()
        return session.query(Station).all()

    def get_all_helper_observable_ids(self):
        session = self.Session()
        session.close()
        return session.query(HelperTemplateIDs).all()

    def get_helper_for_describe_sensor(self, station_id, sensor_id, observable_id):
        session = self.Session()
        exists = session.query(HelperTemplateIDs).filter(
            and_(HelperTemplateIDs.station_id == station_id,
                 HelperTemplateIDs.sensor_id == sensor_id,
                 HelperTemplateIDs.observable_id == observable_id
                 ))

        if exists.first() is not None:
            print(exists.first().observable_id)
            session.expunge_all()
            session.close()
            return exists.first()
        else:
            return None

    def get_observations_by_helper_id(self, helper_id):
        session = self.Session()
        exists = session.query(Observations).filter(Observations.helper_observable_id == helper_id)

        if exists.first() is not None:
            session.expunge_all()
            session.close()
            return exists.all()
        else:
            return None


if __name__ == "__main__":
    x = DatabaseHandler()
    x.__get_station_id_by_tags_station_id__(210)



import copy
import json

import pandas as pd
from sqlalchemy.dialects import sqlite
import numpy as np
from edam.reader.manage import db_connect, db_session
import edam.reader.models as models
from edam.reader.utilities import find_and_describe_templates


class Measurement(object):
    def __init__(self, value, timestamp=None, observable=None, uom=None, station=None, helper=None):
        table = str.maketrans(dict.fromkeys('#*\"'))
        # if str(value).translate(table) == "None":
        #     self.value = "---"
        # else:
        #     self.value = str(value).translate(table)
        self.value = str(value).translate(table)
        self.timestamp = timestamp
        # TODO: Implement those. Observable and uom should be models.AbstractObservable and model.Uoms respectively
        self.observable = observable
        self.uom = uom
        self.station = station
        self.helper = helper
    
    def __repr__(self):
        return self.value


class Station(object):
    def __init__(self, station: models.Station, df: pd.DataFrame):
        self.name = station.name
        self.mobile = station.mobile
        self.location = station.location
        self.latitude = station.latitude
        self.longitude = station.longitude
        self.id = station.id
        
        self.tags = json.loads(station.tags)
        
        self.data = df
    
    def __repr__(self):
        return self.name


class DatabaseHandler(object):
    """"""
    
    def __init__(self):
        """
        Initializes database connection and sessionmaker.
        """
        
        self.engine = db_connect()
        
        self.Session = db_session
    
    def retrieve_stations_data(self, station: models.Station, template_for_lines):
        # print(template_for_lines)
        session = self.Session()
        engine = self.engine
        dict_with_relevant_to_station_help_ids = {helper.id: helper.observable_id for helper in station.helper}
        sql = session.query(models.Observations). \
            filter(models.Observations.helper_observable_id.in_(list(dict_with_relevant_to_station_help_ids)))
        sql_literal = str(sql.statement.compile(dialect=sqlite.dialect(), compile_kwargs={"literal_binds": True}))
        
        df = pd.read_sql_query(sql_literal, engine)
        
        df['helper_observable_id'] = df['helper_observable_id']. \
            apply(lambda x: dict_with_relevant_to_station_help_ids[x])
        df.drop('id', axis=1, inplace=True)
        df = df.pivot_table(index=df.timestamp, columns='helper_observable_id', values='value', aggfunc='first')
        df.reset_index(inplace=True)
        df["timestamp"] = df["timestamp"].apply(lambda x: pd.to_datetime(x))
        df.set_index(keys=["timestamp"], drop=False, inplace=True)
        
        station_df = copy.deepcopy(df)
        
        # Convert values to Measurement objects
        observables_list = list(df)
        observables_list.remove('timestamp')
        table = str.maketrans(dict.fromkeys('#*\"'))
        for observable in observables_list:
            df[observable] = df[observable].apply(lambda x: Measurement(x))
            try:
                station_df[observable] = station_df[observable].apply(lambda x: str(x).translate(table))
            except Exception as e:
                print(observable, e.args)
        # TODO: This is soooooo dangerous. Please re-implement......
        station_df = station_df.replace('None', np.nan)
        # station_df = station_df.fillna('---')
        # print(station_df)
        
        zip_argument = map(lambda x: "df." + x, template_for_lines)
        
        zip_argument = ",".join(zip_argument)
        
        zip_argument = eval("zip(%s)" % zip_argument)
        return Station(station, df=station_df), zip_argument
    
    def retrieve_object_from_id(self, table, object_id):
        session = self.Session()
        item = session.query(getattr(models, table)).filter(getattr(models, table).id == object_id)
        session.close()
        
        return item.first()
    
    @staticmethod
    def retrieve_templates():
        
        return find_and_describe_templates()
    
    def retrieve_stations(self):
        session = self.Session()
        items = session.query(models.Station).all()
        dictionary = dict()
        for station in items:
            dictionary[station.id] = self.station2json(station)
        session.close()
        return dictionary
    
    @staticmethod
    def station2json(station: models.Station):
        tags = json.loads(station.tags)
        dictionary = copy.deepcopy(station.__dict__)
        
        dictionary['tags'] = tags
        
        dictionary['observables'] = dict()
        for helper in station.helper:
            helper: models.HelperTemplateIDs
            dictionary['observables'][helper.observable_id] = dict()
            dictionary['observables'][helper.observable_id]['start_date'] = str(helper.start_date)
            dictionary['observables'][helper.observable_id]['end_date'] = str(helper.end_date)
            dictionary['observables'][helper.observable_id]['observations'] = helper.number_of_observations
            dictionary['observables'][helper.observable_id]['frequency'] = helper.frequency
            dictionary['observables'][helper.observable_id]['observable'] = helper.observable.name
            dictionary['observables'][helper.observable_id]['observed_in'] = helper.uom.name
            # dictionary['observables'][helper.observable.name]['observed_in'] = helper.uom.name
            # dictionary['observables'][helper.observable.name]['observed_in_symbol'] = helper.uom.symbol
            # dictionary['observables'][helper.observable.name]['ontology'] = helper.observable.ontology
            # dictionary['observables'][helper.observable.name]['input_file_path'] = helper.input_file_path
            # dictionary['observables'][helper.observable.name]['observable_id'] = helper.observable_id
        # dictionary.pop('id', None)
        dictionary.pop('helper', None)
        dictionary.pop('_sa_instance_state', None)
        
        return dictionary


if __name__ == "__main__":
    test = DatabaseHandler()
    # test.retrieve_stations()
    # template_for_arguments = "timestamp,radn,maxt,mint,rain,wind,RH"
    # list_template_for_arguments = template_for_arguments.split(',')
    # station_object = test.retrieve_object_from_id(table='Station', object_id=1)  # type: models.Station
    # results = test.retrieve_stations_data(station_object, list_template_for_arguments)
    results = test.retrieve_templates()
    print(results)
    
    # print(results[0].helper.observable_id)



import click

from edam.reader.Workflow import Workflow
from edam.reader.manage import DatabaseInstantiation
from edam.reader.utilities import check_if_path_exists, identify_input_type, determine_storage_type
from edam.viewer.app.views import app
from edam.settings import SERVER
from datetime import datetime


@click.command()
@click.option('--input', required=True, help='input string')
@click.option('--query', required=False, help='SQL query')
@click.option('--template', required=True, help='template file to parse data with')
@click.option('--metadata', required=True, help='configuration file to annotate data with')
@click.option('--var', required=False, default="", help='Extra variables for URI generation')
@click.option('--storage', required=False, default='file', type=click.Choice(['file', 'memory']),
              help="Whether input files to be stored or not")
@click.option('--drop', required=False, default='no', type=click.Choice(['yes', 'no']),
              help="Whether to drop stored data or not")
def cli(input, template, query, metadata, var, storage, drop):
    now = datetime.now()
    if drop == "yes":
        # Dropping database
        DatabaseInstantiation(drop=True)
    else:
        DatabaseInstantiation(drop=False)
    
    template_path, template_object = handle_input_files(template)
    config_path, config_object = handle_input_files(metadata)
    
    success, inputs_path, file_type = identify_input_type(input_file=input, extra_variables=var,
                                                          template=template_path, template_object=template_object,
                                                          sql_query=query,
                                                          storage=determine_storage_type(storage_as_string=storage))
    if success:
        if not inputs_path:
            # It means template and input file does not match
            click.echo(f"I can't match template: {template_path} with input: {input}")
            exit(3)
        elif inputs_path:
            if template_path and config_path:
                mid_time = datetime.now()
                Workflow(input_list=inputs_path, template_file=template_object, configuration_file=config_object)
                end_time = datetime.now()
                print(f"Download all data: {mid_time - now}")
                
                print(f"Store all data: {end_time - mid_time}")
                print(f"Total time: {end_time - now}")
                # run()
            else:
                click.echo("No template or config were given")
        else:
            click.echo(f"{input} does not exist")
            exit(2)
        # run()
    else:
        click.echo(f"{input} does not exist")
        exit(4)


def handle_input_files(filename):
    if filename == "":
        # User didn't provide a template... Are we going to check every template?
        pass
    else:
        exists, file_type, file_path, file_object = check_if_path_exists(filename=filename)
        if exists:
            return file_path, file_object
        else:
            click.echo(f"{filename} does not exist")
            raise SystemExit(0)
            # raise Exception("File does not exit")


def run():
    app.jinja_env.auto_reload = True
    app.config['TEMPLATES_AUTO_RELOAD'] = True
    app.run(host=SERVER['host'], port=SERVER['port'], debug=SERVER['debug'])


if __name__ == '__main__':
    cli()
